# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2017, 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
import struct
import zlib
from binascii import hexlify

from twisted.python.compat import nativeString

from ._util import (
    group_by_topic_and_partition, read_int_string, read_short_ascii,
    read_short_bytes, read_short_text, relative_unpack, write_int_string,
    write_short_ascii, write_short_bytes, write_short_text,
)
from .codec import gzip_decode, gzip_encode, snappy_decode, snappy_encode
from .common import (
    CODEC_GZIP, CODEC_NONE, CODEC_SNAPPY, BrokerMetadata, BufferUnderflowError,
    ChecksumError, ConsumerFetchSizeTooSmall, ConsumerMetadataResponse,
    FetchResponse, InvalidMessageError, Message, OffsetAndMessage,
    OffsetCommitResponse, OffsetFetchResponse, OffsetResponse,
    PartitionMetadata, ProduceResponse, ProtocolError, TopicMetadata,
    UnsupportedCodecError, _HeartbeatResponse, _JoinGroupProtocolMetadata,
    _JoinGroupResponse, _JoinGroupResponseMember, _LeaveGroupResponse,
    _SyncGroupMemberAssignment, _SyncGroupResponse,
)

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

_SUPPORTED_CODECS = (CODEC_GZIP, CODEC_NONE, CODEC_SNAPPY)
ATTRIBUTE_CODEC_MASK = 0x03
MAX_BROKERS = 1024

# Default number of msecs the lead-broker will wait for replics to
# ack produce requests before failing the request
DEFAULT_REPLICAS_ACK_TIMEOUT_MSECS = 1000


class _ReprRequest(object):
    r"""
    Wrapper for request `bytes` that gives it a comprehensible repr for use in
    log messages.

    >>> _ReprRequest(b'\0\x02\0\0\0\0\0\xff')
    ListOffsetsRequest0 correlationId=16 (8 bytes)
    """
    __slots__ = ('_request')

    _struct = struct.Struct('>hhi')

    def __init__(self, request):
        assert isinstance(request, bytes), "request must be bytes, not {!r}".format(request)
        self._request = request

    def __str__(self):
        length = len(self._request)
        if length < 8:
            return 'invalid request ({})'.format(hexlify(self._request).decode('ascii'))

        key, version, correlation_id = self._struct.unpack_from(self._request)
        try:
            key_name = KafkaCodec.key_name(key)
        except KeyError:
            return 'request key={}v{} correlationId={} ({:,d} bytes)'.format(
                key, version, correlation_id, length,
            )

        return '{}Request{} correlationId={} ({:,d} bytes)'.format(
            key_name, version, correlation_id, length,
        )

    def __repr__(self):
        return self.__str__()


class KafkaCodec(object):
    """
    Class to encapsulate all of the protocol encoding/decoding.
    This class does not have any state associated with it, it is purely
    for organization.
    """
    # https://kafka.apache.org/protocol.html#protocol_api_keys
    PRODUCE_KEY = 0
    FETCH_KEY = 1
    OFFSET_KEY = 2
    METADATA_KEY = 3
    # Non-user facing control APIs: 4-7
    OFFSET_COMMIT_KEY = 8
    OFFSET_FETCH_KEY = 9
    CONSUMER_METADATA_KEY = 10  # deprecated
    FIND_COORDINATOR_KEY = 10
    JOIN_GROUP_KEY = 11
    HEARTBEAT_KEY = 12
    LEAVE_GROUP_KEY = 13
    SYNC_GROUP_KEY = 14
    DESCRIBE_GROUP_KEY = 15
    LIST_GROUPS_KEY = 16
    SASL_HANDSHAKE_KEY = 17
    API_VERSIONS_KEY = 18
    CREATE_TOPICS_KEY = 19
    DELETE_TOPICS_KEY = 20

    _key_to_name = {
        PRODUCE_KEY: 'Produce',
        FETCH_KEY: 'Fetch',
        OFFSET_KEY: 'ListOffsets',
        METADATA_KEY: 'Metadata',
        OFFSET_COMMIT_KEY: 'OffsetCommit',
        OFFSET_FETCH_KEY: 'OffsetFetch',
        FIND_COORDINATOR_KEY: 'FindCoordinator',
        JOIN_GROUP_KEY: 'JoinGroup',
        HEARTBEAT_KEY: 'Heartbeat',
        LEAVE_GROUP_KEY: 'LeaveGroup',
        SYNC_GROUP_KEY: 'SyncGroup',
        DESCRIBE_GROUP_KEY: 'DescribeGroup',
        LIST_GROUPS_KEY: 'ListGroups',
        SASL_HANDSHAKE_KEY: 'SaslHandshake',
        API_VERSIONS_KEY: 'ApiVersions',
        CREATE_TOPICS_KEY: 'CreateTopics',
        DELETE_TOPICS_KEY: 'DeleteTopics',
    }

    @classmethod
    def key_name(cls, key):
        return cls._key_to_name[key]

    ###################
    #   Private API   #
    ###################

    @classmethod
    def _encode_message_header(cls, client_id, correlation_id, request_key,
                               api_version=0):
        """
        Encode the common request envelope

        :param bytes client_id: Client identifier, a short bytesting.
        :param int correlation_id: 32-bit int
        :param int request_key: Request type identifier, a 16-bit int. See the
            ``*_KEY`` constants above.
        :param int api_version: Version of the request, defaulting to 0.
        """
        return (struct.pack('>hhih',
                            request_key,          # ApiKey
                            api_version,          # ApiVersion
                            correlation_id,       # CorrelationId
                            len(client_id)) +     # ClientId size
                client_id)                        # ClientId

    @classmethod
    def _encode_message_set(cls, messages, offset=None):
        """
        Encode a MessageSet. Unlike other arrays in the protocol,
        MessageSets are not length-prefixed.  Format::

            MessageSet => [Offset MessageSize Message]
              Offset => int64
              MessageSize => int32
        """
        message_set = []
        incr = 1
        if offset is None:
            incr = 0
            offset = 0
        for message in messages:
            encoded_message = KafkaCodec._encode_message(message)
            message_set.append(struct.pack('>qi', offset, len(encoded_message)))
            message_set.append(encoded_message)
            offset += incr
        return b''.join(message_set)

    @classmethod
    def _encode_message(cls, message):
        """
        Encode a single message.

        The magic number of a message is a format version number.  The only
        supported magic number right now is zero.  Format::

            Message => Crc MagicByte Attributes Key Value
              Crc => int32
              MagicByte => int8
              Attributes => int8
              Key => bytes
              Value => bytes
        """
        if message.magic == 0:
            msg = struct.pack('>BB', message.magic, message.attributes)
            msg += write_int_string(message.key)
            msg += write_int_string(message.value)
            crc = zlib.crc32(msg) & 0xffffffff  # Ensure unsigned
            msg = struct.pack('>I', crc) + msg
        else:
            raise ProtocolError("Unexpected magic number: %d" % message.magic)
        return msg

    @classmethod
    def _decode_message_set_iter(cls, data):
        """
        Iteratively decode a MessageSet

        Reads repeated elements of (offset, message), calling decode_message
        to decode a single message. Since compressed messages contain futher
        MessageSets, these two methods have been decoupled so that they may
        recurse easily.
        """
        cur = 0
        read_message = False
        while cur < len(data):
            try:
                ((offset, ), cur) = relative_unpack('>q', data, cur)
                (msg, cur) = read_int_string(data, cur)
                msgIter = KafkaCodec._decode_message(msg, offset)
                for (offset, message) in msgIter:
                    read_message = True
                    yield OffsetAndMessage(offset, message)
            except BufferUnderflowError:
                # NOTE: Not sure this is correct error handling:
                # Is it possible to get a BUE if the message set is somewhere
                # in the middle of the fetch response? If so, we probably have
                # an issue that's not fetch size too small.
                # Aren't we ignoring errors if we fail to unpack data by
                # raising StopIteration()?
                # If _decode_message() raises a ChecksumError, couldn't that
                # also be due to the fetch size being too small?
                if read_message is False:
                    # If we get a partial read of a message, but haven't
                    # yielded anything there's a problem
                    raise ConsumerFetchSizeTooSmall()
                else:
                    return

    @classmethod
    def _decode_message(cls, data, offset):
        """
        Decode a single Message

        The only caller of this method is decode_message_set_iter.
        They are decoupled to support nested messages (compressed MessageSets).
        The offset is actually read from decode_message_set_iter (it is part
        of the MessageSet payload).
        """
        ((crc, magic, att), cur) = relative_unpack('>IBB', data, 0)
        if crc != zlib.crc32(data[4:]) & 0xffffffff:
            raise ChecksumError("Message checksum failed")

        (key, cur) = read_int_string(data, cur)
        (value, cur) = read_int_string(data, cur)

        codec = att & ATTRIBUTE_CODEC_MASK

        if codec == CODEC_NONE:
            yield (offset, Message(magic, att, key, value))

        elif codec == CODEC_GZIP:
            gz = gzip_decode(value)
            for (offset, msg) in KafkaCodec._decode_message_set_iter(gz):
                yield (offset, msg)

        elif codec == CODEC_SNAPPY:
            snp = snappy_decode(value)
            for (offset, msg) in KafkaCodec._decode_message_set_iter(snp):
                yield (offset, msg)

        else:
            raise ProtocolError('Unsupported codec 0b{:b}'.format(codec))

    ##################
    #   Public API   #
    ##################

    @classmethod
    def get_response_correlation_id(cls, data):
        """
        return just the correlationId part of the response

        :param bytes data: bytes to decode
        """
        ((correlation_id,), cur) = relative_unpack('>i', data, 0)
        return correlation_id

    @classmethod
    def encode_produce_request(cls, client_id, correlation_id,
                               payloads=None, acks=1,
                               timeout=DEFAULT_REPLICAS_ACK_TIMEOUT_MSECS):
        """
        Encode some ProduceRequest structs

        :param bytes client_id:
        :param int correlation_id:
        :param list payloads: list of ProduceRequest
        :param int acks:

            How "acky" you want the request to be:

            0: immediate response
            1: written to disk by the leader
            2+: waits for this many number of replicas to sync
            -1: waits for all replicas to be in sync

        :param int timeout:
            Maximum time the server will wait for acks from replicas.  This is
            _not_ a socket timeout.
        """
        if not isinstance(client_id, bytes):
            raise TypeError('client_id={!r} should be bytes'.format(client_id))
        payloads = [] if payloads is None else payloads
        grouped_payloads = group_by_topic_and_partition(payloads)

        message = cls._encode_message_header(client_id, correlation_id,
                                             KafkaCodec.PRODUCE_KEY)

        message += struct.pack('>hii', acks, timeout, len(grouped_payloads))

        for topic, topic_payloads in grouped_payloads.items():
            message += write_short_ascii(topic)

            message += struct.pack('>i', len(topic_payloads))
            for partition, payload in topic_payloads.items():
                msg_set = KafkaCodec._encode_message_set(payload.messages)
                message += struct.pack('>ii', partition, len(msg_set))
                message += msg_set

        return message

    @classmethod
    def decode_produce_response(cls, data):
        """
        Decode bytes to a ProduceResponse

        :param bytes data: bytes to decode
        :returns: iterable of `afkak.common.ProduceResponse`
        """
        ((correlation_id, num_topics), cur) = relative_unpack('>ii', data, 0)

        for _i in range(num_topics):
            topic, cur = read_short_ascii(data, cur)
            ((num_partitions,), cur) = relative_unpack('>i', data, cur)
            for _i in range(num_partitions):
                ((partition, error, offset), cur) = relative_unpack('>ihq', data, cur)

                yield ProduceResponse(topic, partition, error, offset)

    @classmethod
    def encode_fetch_request(cls, client_id, correlation_id, payloads=None,
                             max_wait_time=100, min_bytes=4096):
        """
        Encodes some FetchRequest structs

        :param bytes client_id:
        :param int correlation_id:
        :param list payloads: list of :class:`FetchRequest`
        :param int max_wait_time: how long to block waiting on min_bytes of data
        :param int min_bytes:
            the minimum number of bytes to accumulate before returning the
            response
        """
        payloads = [] if payloads is None else payloads
        grouped_payloads = group_by_topic_and_partition(payloads)

        message = cls._encode_message_header(client_id, correlation_id,
                                             KafkaCodec.FETCH_KEY)

        assert isinstance(max_wait_time, int)

        # -1 is the replica id
        message += struct.pack('>iiii', -1, max_wait_time, min_bytes,
                               len(grouped_payloads))

        for topic, topic_payloads in grouped_payloads.items():
            message += write_short_ascii(topic)
            message += struct.pack('>i', len(topic_payloads))
            for partition, payload in topic_payloads.items():
                message += struct.pack('>iqi', partition, payload.offset,
                                       payload.max_bytes)

        return message

    @classmethod
    def decode_fetch_response(cls, data):
        """
        Decode bytes to a FetchResponse

        :param bytes data: bytes to decode
        """
        ((correlation_id, num_topics), cur) = relative_unpack('>ii', data, 0)

        for _i in range(num_topics):
            (topic, cur) = read_short_ascii(data, cur)
            ((num_partitions,), cur) = relative_unpack('>i', data, cur)

            for _i in range(num_partitions):
                ((partition, error, highwater_mark_offset), cur) = \
                    relative_unpack('>ihq', data, cur)

                (message_set, cur) = read_int_string(data, cur)

                yield FetchResponse(
                    topic, partition, error,
                    highwater_mark_offset,
                    KafkaCodec._decode_message_set_iter(message_set))

    @classmethod
    def encode_offset_request(cls, client_id, correlation_id, payloads=None):
        payloads = [] if payloads is None else payloads
        grouped_payloads = group_by_topic_and_partition(payloads)

        message = cls._encode_message_header(client_id, correlation_id,
                                             KafkaCodec.OFFSET_KEY)

        # -1 is the replica id
        message += struct.pack('>ii', -1, len(grouped_payloads))

        for topic, topic_payloads in grouped_payloads.items():
            message += write_short_ascii(topic)
            message += struct.pack('>i', len(topic_payloads))

            for partition, payload in topic_payloads.items():
                message += struct.pack('>iqi', partition, payload.time,
                                       payload.max_offsets)

        return message

    @classmethod
    def decode_offset_response(cls, data):
        """
        Decode bytes to an :class:`OffsetResponse`

        :param bytes data: bytes to decode
        """
        ((correlation_id, num_topics), cur) = relative_unpack('>ii', data, 0)

        for _i in range(num_topics):
            (topic, cur) = read_short_ascii(data, cur)
            ((num_partitions,), cur) = relative_unpack('>i', data, cur)

            for _i in range(num_partitions):
                ((partition, error, num_offsets), cur) = \
                    relative_unpack('>ihi', data, cur)

                offsets = []
                for _i in range(num_offsets):
                    ((offset,), cur) = relative_unpack('>q', data, cur)
                    offsets.append(offset)

                yield OffsetResponse(topic, partition, error, tuple(offsets))

    @classmethod
    def encode_metadata_request(cls, client_id, correlation_id, topics=None):
        """
        Encode a MetadataRequest

        :param bytes client_id: string
        :param int correlation_id: int
        :param list topics: list of text
        """
        topics = [] if topics is None else topics
        message = [
            cls._encode_message_header(client_id, correlation_id,
                                       KafkaCodec.METADATA_KEY),
            struct.pack('>i', len(topics)),
        ]
        for topic in topics:
            message.append(write_short_ascii(topic))
        return b''.join(message)

    @classmethod
    def decode_metadata_response(cls, data):
        """
        Decode bytes to a MetadataResponse

        :param bytes data: bytes to decode
        """
        ((correlation_id, numbrokers), cur) = relative_unpack('>ii', data, 0)

        # In testing, I saw this routine swap my machine to death when
        # passed bad data. So, some checks are in order...
        if numbrokers > MAX_BROKERS:
            raise InvalidMessageError(
                "Brokers:{} exceeds max:{}".format(numbrokers, MAX_BROKERS))

        # Broker info
        brokers = {}
        for _i in range(numbrokers):
            ((nodeId, ), cur) = relative_unpack('>i', data, cur)
            (host, cur) = read_short_ascii(data, cur)
            ((port,), cur) = relative_unpack('>i', data, cur)
            brokers[nodeId] = BrokerMetadata(nodeId, nativeString(host), port)

        # Topic info
        ((num_topics,), cur) = relative_unpack('>i', data, cur)
        topic_metadata = {}

        for _i in range(num_topics):
            ((topic_error,), cur) = relative_unpack('>h', data, cur)
            (topic_name, cur) = read_short_ascii(data, cur)
            ((num_partitions,), cur) = relative_unpack('>i', data, cur)
            partition_metadata = {}

            for _j in range(num_partitions):
                ((partition_error_code, partition, leader, numReplicas),
                 cur) = relative_unpack('>hiii', data, cur)

                (replicas, cur) = relative_unpack(
                    '>%di' % numReplicas, data, cur)

                ((num_isr,), cur) = relative_unpack('>i', data, cur)
                (isr, cur) = relative_unpack('>%di' % num_isr, data, cur)

                partition_metadata[partition] = \
                    PartitionMetadata(
                        topic_name, partition, partition_error_code, leader,
                        replicas, isr)

            topic_metadata[topic_name] = TopicMetadata(
                topic_name, topic_error, partition_metadata)

        return brokers, topic_metadata

    @classmethod
    def encode_consumermetadata_request(cls, client_id, correlation_id,
                                        consumer_group):
        """
        Encode a ConsumerMetadataRequest

        :param bytes client_id: string
        :param int correlation_id: int
        :param str consumer_group: string
        """
        message = cls._encode_message_header(client_id, correlation_id,
                                             KafkaCodec.CONSUMER_METADATA_KEY)
        message += write_short_ascii(consumer_group)
        return message

    @classmethod
    def decode_consumermetadata_response(cls, data):
        """
        Decode bytes to a ConsumerMetadataResponse

        :param bytes data: bytes to decode
        """
        (correlation_id, error_code, node_id), cur = \
            relative_unpack('>ihi', data, 0)
        host, cur = read_short_ascii(data, cur)
        (port,), cur = relative_unpack('>i', data, cur)

        return ConsumerMetadataResponse(
            error_code, node_id, nativeString(host), port)

    @classmethod
    def encode_offset_commit_request(cls, client_id, correlation_id,
                                     group, group_generation_id, consumer_id,
                                     payloads):
        """
        Encode some OffsetCommitRequest structs (v1)

        :param bytes client_id: string
        :param int correlation_id: int
        :param str group: the consumer group to which you are committing offsets
        :param int group_generation_id: int32, generation ID of the group
        :param str consumer_id: string, Identifier for the consumer
        :param list payloads: list of :class:`OffsetCommitRequest`
        """
        assert consumer_id is not None
        grouped_payloads = group_by_topic_and_partition(payloads)

        message = cls._encode_message_header(
            client_id, correlation_id, KafkaCodec.OFFSET_COMMIT_KEY,
            api_version=1,
        )

        message += write_short_ascii(group)
        message += struct.pack('>i', group_generation_id)
        message += write_short_ascii(consumer_id)
        message += struct.pack('>i', len(grouped_payloads))

        for topic, topic_payloads in grouped_payloads.items():
            message += write_short_ascii(topic)
            message += struct.pack('>i', len(topic_payloads))

            for partition, payload in topic_payloads.items():
                message += struct.pack('>iqq', partition, payload.offset,
                                       payload.timestamp)
                message += write_short_bytes(payload.metadata)

        return message

    @classmethod
    def decode_offset_commit_response(cls, data):
        """
        Decode bytes to an OffsetCommitResponse

        :param bytes data: bytes to decode
        """
        ((correlation_id,), cur) = relative_unpack('>i', data, 0)
        ((num_topics,), cur) = relative_unpack('>i', data, cur)

        for _i in range(num_topics):
            (topic, cur) = read_short_ascii(data, cur)
            ((num_partitions,), cur) = relative_unpack('>i', data, cur)

            for _i in range(num_partitions):
                ((partition, error), cur) = relative_unpack('>ih', data, cur)
                yield OffsetCommitResponse(topic, partition, error)

    @classmethod
    def encode_offset_fetch_request(cls, client_id, correlation_id,
                                    group, payloads):
        """
        Encode some OffsetFetchRequest structs

        :param bytes client_id: string
        :param int correlation_id: int
        :param bytes group: string, the consumer group you are fetching offsets for
        :param list payloads: list of :class:`OffsetFetchRequest`
        """
        grouped_payloads = group_by_topic_and_partition(payloads)
        message = cls._encode_message_header(
            client_id, correlation_id, KafkaCodec.OFFSET_FETCH_KEY,
            api_version=1)

        message += write_short_ascii(group)
        message += struct.pack('>i', len(grouped_payloads))

        for topic, topic_payloads in grouped_payloads.items():
            message += write_short_ascii(topic)
            message += struct.pack('>i', len(topic_payloads))

            for partition in topic_payloads:
                message += struct.pack('>i', partition)

        return message

    @classmethod
    def decode_offset_fetch_response(cls, data):
        """
        Decode bytes to an OffsetFetchResponse

        :param bytes data: bytes to decode
        """

        ((correlation_id,), cur) = relative_unpack('>i', data, 0)
        ((num_topics,), cur) = relative_unpack('>i', data, cur)

        for _i in range(num_topics):
            (topic, cur) = read_short_ascii(data, cur)
            ((num_partitions,), cur) = relative_unpack('>i', data, cur)

            for _i in range(num_partitions):
                ((partition, offset), cur) = relative_unpack('>iq', data, cur)
                (metadata, cur) = read_short_bytes(data, cur)
                ((error,), cur) = relative_unpack('>h', data, cur)

                yield OffsetFetchResponse(topic, partition, offset,
                                          metadata, error)

    @classmethod
    def encode_join_group_request(cls, client_id, correlation_id,
                                  payload):
        """
        Encode a JoinGroupRequest

        :param bytes client_id: string
        :param int correlation_id: int
        :param :class:`JoinGroupRequest` payload: payload
        """
        message = cls._encode_message_header(
            client_id, correlation_id, KafkaCodec.JOIN_GROUP_KEY,
            api_version=0)

        message += write_short_text(payload.group)
        message += struct.pack('>i', payload.session_timeout)
        message += write_short_text(payload.member_id)
        message += write_short_text(payload.protocol_type)

        message += struct.pack('>i', len(payload.group_protocols))
        for group_protocol in payload.group_protocols:
            message += write_short_ascii(group_protocol.protocol_name)
            message += write_int_string(group_protocol.protocol_metadata)

        return message

    @classmethod
    def encode_join_group_protocol_metadata(cls, version, subscriptions, user_data):
        message = struct.pack('>hi', version, len(subscriptions))
        for subscription in subscriptions:
            message += write_short_text(subscription)
        message += write_int_string(user_data)
        return message

    @classmethod
    def decode_join_group_protocol_metadata(cls, data):
        """
        Decode bytes to a JoinGroupProtocolMetadata

        :param bytes data: bytes to decode
        """
        ((version, num_subscriptions), cur) = relative_unpack('>hi', data, 0)
        subscriptions = []
        for _i in range(num_subscriptions):
            (subscription, cur) = read_short_text(data, cur)
            subscriptions.append(subscription)
        (user_data, cur) = read_int_string(data, cur)
        return _JoinGroupProtocolMetadata(version, subscriptions, user_data)

    @classmethod
    def decode_join_group_response(cls, data):
        """
        Decode bytes to a JoinGroupResponse

        :param bytes data: bytes to decode
        """

        ((correlation_id, error, generation_id), cur) = relative_unpack('>ihi', data, 0)
        (group_protocol, cur) = read_short_text(data, cur)
        (leader_id, cur) = read_short_text(data, cur)
        (member_id, cur) = read_short_text(data, cur)
        ((num_members,), cur) = relative_unpack('>i', data, cur)

        members = []
        for _i in range(num_members):
            (response_member_id, cur) = read_short_text(data, cur)
            (response_member_data, cur) = read_int_string(data, cur)
            members.append(_JoinGroupResponseMember(response_member_id, response_member_data))
        return _JoinGroupResponse(error, generation_id, group_protocol,
                                  leader_id, member_id, members)

    @classmethod
    def encode_leave_group_request(cls, client_id, correlation_id, payload):
        """
        Encode a LeaveGroupRequest
        :param bytes client_id: string
        :param int correlation_id: int
        :param :class:`LeaveGroupRequest` payload: payload
        """
        message = cls._encode_message_header(
            client_id, correlation_id, KafkaCodec.LEAVE_GROUP_KEY,
            api_version=0)

        message += write_short_text(payload.group)
        message += write_short_text(payload.member_id)
        return message

    @classmethod
    def decode_leave_group_response(cls, data):
        """
        Decode bytes to a LeaveGroupResponse

        :param bytes data: bytes to decode
        """
        ((correlation_id, error), cur) = relative_unpack('>ih', data, 0)
        return _LeaveGroupResponse(error)

    @classmethod
    def encode_heartbeat_request(cls, client_id, correlation_id, payload):
        """
        Encode a HeartbeatRequest

        :param bytes client_id: string
        :param int correlation_id: int
        :param :class:`_HeartbeatRequest` payload: payload
        """
        message = cls._encode_message_header(
            client_id, correlation_id, KafkaCodec.HEARTBEAT_KEY,
            api_version=0)

        message += write_short_text(payload.group)
        message += struct.pack('>i', payload.generation_id)
        message += write_short_text(payload.member_id)
        return message

    @classmethod
    def decode_heartbeat_response(cls, data):
        """
        Decode bytes to a `_HeartbeatResponse`

        :param bytes data: bytes to decode
        """
        ((correlation_id, error), cur) = relative_unpack('>ih', data, 0)
        return _HeartbeatResponse(error)

    @classmethod
    def encode_sync_group_request(cls, client_id, correlation_id, payload):
        """
        Encode a SyncGroupRequest

        :param bytes client_id: string
        :param int correlation_id: int
        :param payload: :class:`_SyncGroupRequest`
        """
        message = cls._encode_message_header(
            client_id, correlation_id, KafkaCodec.SYNC_GROUP_KEY,
            api_version=0)

        message += write_short_text(payload.group)
        message += struct.pack('>i', payload.generation_id)
        message += write_short_text(payload.member_id)

        message += struct.pack('>i', len(payload.group_assignment))
        for assignment in payload.group_assignment:
            message += write_short_text(assignment.member_id)
            message += write_int_string(assignment.member_metadata)

        return message

    @classmethod
    def decode_sync_group_response(cls, data):
        """
        Decode bytes to a SyncGroupResponse

        :param bytes data: bytes to decode
        """
        ((correlation_id, error), cur) = relative_unpack('>ih', data, 0)
        (member_assignment, cur) = read_int_string(data, cur)
        return _SyncGroupResponse(error, member_assignment)

    @classmethod
    def encode_sync_group_member_assignment(cls, version, assignments, user_data):
        message = struct.pack('>h', version)
        message += struct.pack('>i', len(assignments))
        for topic, partitions in assignments.items():
            message += write_short_ascii(topic)
            message += struct.pack('>i%si' % len(partitions), len(partitions), *partitions)
        message += write_int_string(user_data)
        return message

    @classmethod
    def decode_sync_group_member_assignment(cls, data):
        """
        Decode bytes to a SyncGroupMemberAssignment

        :param bytes data: bytes to decode
        """

        ((version, num_assignments), cur) = relative_unpack('>hi', data, 0)
        if version != 0:
            raise ProtocolError('Unsupported SyncGroupMemberAssignment version {}'.format(version))
        assignments = {}
        for _i in range(num_assignments):
            (topic, cur) = read_short_ascii(data, cur)
            ((num_partitions,), cur) = relative_unpack('>i', data, cur)
            (partitions, cur) = relative_unpack('>%si' % num_partitions, data, cur)
            assignments[topic] = partitions
        (user_data, cur) = read_int_string(data, cur)
        return _SyncGroupMemberAssignment(version, assignments, user_data)


def create_message(payload, key=None):
    """
    Construct a :class:`Message`

    :param payload: The payload to send to Kafka.
    :type payload: :class:`bytes` or ``None``
    :param key: A key used to route the message when partitioning and to
        determine message identity on a compacted topic.
    :type key: :class:`bytes` or ``None``
    """
    assert payload is None or isinstance(payload, bytes), 'payload={!r} should be bytes or None'.format(payload)
    assert key is None or isinstance(key, bytes), 'key={!r} should be bytes or None'.format(key)
    return Message(0, 0, key, payload)


def create_gzip_message(message_set):
    """
    Construct a gzip-compressed message containing multiple messages

    The given messages will be encoded, compressed, and sent as a single atomic
    message to Kafka.

    :param list message_set: a list of :class:`Message` instances
    """
    encoded_message_set = KafkaCodec._encode_message_set(message_set)

    gzipped = gzip_encode(encoded_message_set)
    return Message(0, CODEC_GZIP, None, gzipped)


def create_snappy_message(message_set):
    """
    Construct a Snappy-compressed message containing multiple messages

    The given messages will be encoded, compressed, and sent as a single atomic
    message to Kafka.

    :param list message_set: a list of :class:`Message` instances
    """
    encoded_message_set = KafkaCodec._encode_message_set(message_set)
    snapped = snappy_encode(encoded_message_set)
    return Message(0, CODEC_SNAPPY, None, snapped)


def create_message_set(requests, codec=CODEC_NONE):
    """
    Create a message set from a list of requests.

    Each request can have a list of messages and its own key.  If codec is
    :data:`CODEC_NONE`, return a list of raw Kafka messages. Otherwise, return
    a list containing a single codec-encoded message.

    :param codec:
        The encoding for the message set, one of the constants:

        - `afkak.CODEC_NONE`
        - `afkak.CODEC_GZIP`
        - `afkak.CODEC_SNAPPY`

    :raises: :exc:`UnsupportedCodecError` for an unsupported codec
    """
    msglist = []
    for req in requests:
        msglist.extend([create_message(m, key=req.key) for m in req.messages])

    if codec == CODEC_NONE:
        return msglist
    elif codec == CODEC_GZIP:
        return [create_gzip_message(msglist)]
    elif codec == CODEC_SNAPPY:
        return [create_snappy_message(msglist)]
    else:
        raise UnsupportedCodecError("Codec 0x%02x unsupported" % codec)



# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2016, 2017, 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import functools
import logging
import os
import sys
import time
import uuid
from pprint import pformat

from twisted.internet.defer import inlineCallbacks, returnValue

from afkak import KafkaClient
from afkak.common import (
    OffsetRequest, PartitionUnavailableError, RetriableBrokerResponseError,
    SendRequest, TopicAndPartition,
)
from afkak.test.int.fixtures import KafkaHarness
from afkak.test.testutil import async_delay, random_string

log = logging.getLogger(__name__)

__all__ = [
    'IntegrationMixin',
    'kafka_versions',
    'stat',
]


def stat(key, value):
    print("##teamcity[buildStatisticValue key='{}' value='{}']".format(
        key, value), file=sys.stderr)


def make_send_requests(msgs, topic=None, key=None):
    return [SendRequest(topic, key, msgs, None)]


def kafka_versions(*versions):
    def kafka_versions(func):
        @functools.wraps(func)
        def wrapper(self):
            kafka_version = os.environ.get('KAFKA_VERSION')

            if not kafka_version:
                self.skipTest("no kafka version specified")  # pragma: no cover
            elif 'all' not in versions and kafka_version not in versions:
                self.skipTest("unsupported kafka version")  # pragma: no cover

            return func(self)
        return wrapper
    return kafka_versions


@inlineCallbacks
def ensure_topic_creation(client, topic_name, fully_replicated=True, timeout=5):
    '''
    With the default Kafka configuration, just querying for the metadata
    for a particular topic will auto-create that topic.

    :param client: `afkak.client.KafkaClient` instance

    :param str topic_name: Topic name

    :param bool fully_replicated:
        If ``True``, check whether all partitions for the topic have been
        assigned brokers. This doesn't ensure that producing to the topic will
        succeed, though—there is a window after the partition is assigned
        before the broker can actually accept writes. In this case the broker
        will respond with a retriable error (see
        `IntegrationMixin.retry_broker_errors()`).

        If ``False``, only check that any metadata exists for the topic.

    :param timeout: Number of seconds to wait.
    '''
    start_time = time.time()
    if fully_replicated:
        check_func = client.topic_fully_replicated
    else:
        check_func = client.has_metadata_for_topic
    yield client.load_metadata_for_topics(topic_name)

    def topic_info():
        if topic_name in client.topic_partitions:
            return "Topic {} exists. Partition metadata: {}".format(
                topic_name, pformat([client.partition_meta[TopicAndPartition(topic_name, part)]
                                     for part in client.topic_partitions[topic_name]]),
            )
        else:
            return "No metadata for topic {} found.".format(topic_name)

    while not check_func(topic_name):
        yield async_delay(clock=client.reactor)
        if time.time() > start_time + timeout:
            raise Exception((
                "Timed out waiting topic {} creation after {} seconds. {}"
            ).format(topic_name, timeout, topic_info()))
        else:
            log.debug('Still waiting topic creation: %s.', topic_info())
        yield client.load_metadata_for_topics(topic_name)
    log.info('%s', topic_info())


class IntegrationMixin(object):
    """
    Mixin for tests that require a Kafka cluster.

    The `setUp()` and `tearDown()` methods bring up a Kafka cluster and create
    a topic for the test to use.

    Mix this into a subclass of `twisted.trial.unittest.TestCase`. Note that
    you must override *harness_kw* in the subclass.

    :data dict harness_kw:
        Keyword arguments for `harness`. Subclasses must set this to specify
        ``replicas`` (the number of Kafka brokers) and may specify other
        arguments — see `afkak.fixtures.KafkaHarness.start()`.

    :data dict client_kw:
        Keyword arguments for `client`. Subclasses may inject keyword arguments
        by overriding this. The default is empty.

    :ivar str topic:
        Kafka topic name. This may be set in subclasses. If ``None``, a random
        topic name is generated by `setUp()`.

    :ivar harness:
        `afkak.test.fixtures.KafkaHarness` instance. This is created by the
        `setUp()` method and automatically torn down.

    :ivar client:
        `afkak.KafkaClient` instance created by the `setUp()` method.

    :ivar reactor: Twisted reactor.
    """
    topic = None
    from twisted.internet import reactor
    client_kw = {}

    if not os.environ.get('KAFKA_VERSION'):  # pragma: no cover
        skip = 'KAFKA_VERSION is not set'

    @inlineCallbacks
    def setUp(self):
        log.info("Setting up test %s", self.id())

        self.harness = KafkaHarness.start(**self.harness_kw)
        self.addCleanup(self.harness.halt)

        if not self.topic:
            self.topic = "%s-%s" % (
                self.id()[self.id().rindex(".") + 1:], random_string(10))

        self.client = KafkaClient(
            self.harness.bootstrap_hosts,
            clientId=self.__class__.__name__,
            **self.client_kw
        )
        self.addCleanup(self.client.close)

        yield ensure_topic_creation(self.client, self.topic,
                                    fully_replicated=True)

        self._messages = {}

    def tearDown(self):
        log.info("Tearing down test: %r", self)

    @inlineCallbacks
    def current_offset(self, topic, partition):
        offsets, = yield self.client.send_offset_request(
            [OffsetRequest(topic, partition, -1, 1)])
        returnValue(offsets.offsets[0])

    @inlineCallbacks
    def retry_while_broker_errors(self, f, *a, **kw):
        """
        Call a function, retrying on retriable broker errors.

        If calling the function fails with one of these exception types it is
        called again after a short delay:

        * `afkak.common.RetriableBrokerResponseError` (or a subclass thereof)
        * `afkak.common.PartitionUnavailableError`

        The net effect is to keep trying until topic auto-creation completes.

        :param f: callable, which may return a `Deferred`
        :param a: arbitrary positional arguments
        :param kw: arbitrary keyword arguments
        """
        while True:
            try:
                returnValue((yield f(*a, **kw)))
                break
            except (RetriableBrokerResponseError, PartitionUnavailableError):
                yield async_delay(0.1, clock=self.reactor)

    def msg(self, s):
        if s not in self._messages:
            self._messages[s] = (u'%s-%s-%s' % (s, self.id(), uuid.uuid4())).encode('utf-8')

        return self._messages[s]



# -*- coding: utf-8 -*-
# Copyright 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from collections import defaultdict
from datetime import datetime, timedelta

import attr
from mock import Mock
from twisted.internet.defer import (
    Deferred, DeferredQueue, inlineCallbacks, returnValue,
)
from twisted.trial import unittest

from afkak import ConsumerGroup, KafkaClient, create_message
from afkak._group import Coordinator
from afkak.common import ProduceRequest
from afkak.test.int.intutil import IntegrationMixin, kafka_versions
from afkak.test.testutil import async_delay

log = logging.getLogger(__name__)


@attr.s(frozen=True)
class Deadline(object):
    timeout = attr.ib(default=timedelta(seconds=30))
    start = attr.ib(default=attr.Factory(datetime.utcnow))

    def check(self):
        """
        Has the deadline expired?

        :raises AssertionError: when the deadline has expired
        """
        elapsed = datetime.utcnow() - self.start
        if elapsed > self.timeout:
            raise AssertionError("Deadline exceeded: {} > {}".format(
                self.elapsed, self.timeout))


def assert_assignments(topic, partitions, members):
    """
    Are the partitions of a topic properly distributed among the members of the
    group? Specifically:

     -  The expected number of partitions has been assigned.

     -  Partitions are evenly distributed among group members. Some members may
        have an extra partition if the number of partitions doesn't evenly
        divide by the number of members.

     -  Each member's partition assignments are disjoint (each partition is
        assigned to exactly one member).

    :param str topic: Name of the topic to inspect.

    :param int partitions: Number of partitions in the topic.

    :param members: One or more `afkak.ConsumerGroup` instances. Each
        must be configured to consume *topic*.

    :raises AssertionError: when assignments fail validation

    :raises ValueError: when a consumer group isn't consuming the topic at all
    """
    min_count, mod = divmod(partitions, len(members))
    if mod > 0:
        max_count = min_count + 1
    else:
        max_count = min_count

    summary = [
        'topic {!r} has {} partitions'.format(topic, partitions),
        'members should own at least {} partitions, no more than {}'.format(min_count, max_count),
    ]
    failures = []
    partition_to_members = defaultdict(list)

    for member in members:
        if topic not in member.topics:
            raise ValueError("ConsumerGroup {!r} isn't consuming topic {!r}".format(
                member, topic))

        a = set(c.partition for c in member.consumers.get(topic, []))
        summary.append('member {} owns partitions {}'.format(
            member, ' '.join(str(p) for p in sorted(a))))

        if len(a) < min_count:
            failures.append('too few partitions: {}'.format(member))
        elif len(a) > max_count:
            failures.append('too many partitions: {}'.format(member))

        for partition in a:
            partition_to_members[partition].append(repr(member))

    for partition, assignees in partition_to_members.items():
        if len(assignees) > 1:
            failures.append('partition {} assignment collision: {}'.format(
                partition, ' '.join(assignees)))

    if len(partition_to_members) != partitions:
        failures.append('{} partitions assigned, but expected {}'.format(
            len(partition_to_members), partitions))

    if failures:
        summary.append('\nFailed validation:')
        summary.extend(failures)
    message = '\n'.join(summary)
    log.debug(message)
    if failures:
        raise AssertionError(message)


def wait_for_assignments(topic, partitions, members):
    deadline = Deadline()
    while True:
        try:
            assert_assignments(
                topic,
                partitions,
                members,
            )
            break
        except AssertionError:
            deadline.check()
            yield async_delay(0.5)


class TestAfkakGroupIntegration(IntegrationMixin, unittest.TestCase):
    num_partitions = 6
    harness_kw = dict(
        replicas=1,
        partitions=num_partitions,
    )

    @inlineCallbacks
    def send_messages(self, partition, messages):
        log.debug("send_messages(%d, %r)", partition, messages)
        messages = [create_message(self.msg(str(msg))) for msg in messages]
        produce = ProduceRequest(self.topic, partition, messages=messages)
        [resp] = yield self.client.send_produce_request([produce])

        self.assertEqual(resp.error, 0)

        returnValue([x.value for x in messages])

    def when_called(self, object, method):
        """
            returns a Deferred that will be called back after the
            next time the method gets called
        """
        de = Deferred()
        original = getattr(object, method)

        def side_effect(*args, **kwargs):
            ret = original(*args, **kwargs)
            if not de.called:
                de.callback(None)
            return ret
        setattr(object, method, Mock(side_effect=side_effect))
        return de

    @kafka_versions("all")
    @inlineCallbacks
    def test_single_coordinator_join(self):
        coord = Coordinator(self.client, self.id(), ["test-topic"])

        de = self.when_called(coord, 'on_join_complete')
        coord.start()
        yield de
        yield coord.stop()

    @kafka_versions("all")
    @inlineCallbacks
    def test_three_coordinator_join(self):
        self.client2 = KafkaClient(
            self.harness.bootstrap_hosts,
            clientId=self.topic + '2')
        self.addCleanup(self.client2.close)

        self.client3 = KafkaClient(
            self.harness.bootstrap_hosts,
            clientId=self.topic + '3')
        self.addCleanup(self.client3.close)
        coords = [
            Coordinator(
                client, self.id(), topics=["test-topic"],
                retry_backoff_ms=100, heartbeat_interval_ms=100,
                fatal_backoff_ms=1000)
            for client in [self.client, self.client2, self.client3]
        ]
        coords[0].on_join_complete = lambda *args: a_joined.callback(*args)
        coords[0].start()

        a_joined = Deferred()

        # startup the first member of the group
        a_assignment = yield a_joined
        log.warn("first is in")

        self.assertNotEqual(coords[0].generation_id, None)
        self.assertEqual(coords[0].leader_id, coords[0].member_id)
        self.assertEqual(a_assignment, {"test-topic": (0, 1, 2, 3, 4, 5)})
        first_generation_id = coords[0].generation_id

        # now bring someone else into the group
        a_joined, b_joined = Deferred(), Deferred()

        coords[0].on_join_complete = lambda *args: a_joined.callback(*args)
        coords[1].on_join_complete = lambda *args: b_joined.callback(*args)
        log.warn("bringing in second")
        coords[1].start()
        log.warn("waiting for a_joined")
        yield a_joined
        log.warn("waiting for b_joined")
        yield b_joined

        self.assertEqual(coords[0].generation_id, coords[1].generation_id)
        self.assertEqual(coords[0].leader_id, coords[1].leader_id)
        self.assertNotEqual(coords[0].member_id, coords[1].member_id)
        self.assertNotEqual(coords[0].generation_id, first_generation_id)

        # and then bring in a third
        a_joined, b_joined, c_joined = Deferred(), Deferred(), Deferred()

        coords[0].on_join_complete = lambda *args: a_joined.callback(*args)
        coords[1].on_join_complete = lambda *args: b_joined.callback(*args)
        coords[2].on_join_complete = lambda *args: c_joined.callback(*args)
        log.warn("bringing in third")
        coords[2].start()
        log.warn("waiting for a_joined")
        a_assignment = yield a_joined
        log.warn("waiting for b_joined")
        b_assignment = yield b_joined
        log.warn("waiting for c_joined")
        c_assignment = yield c_joined

        self.assertTrue(len(a_assignment["test-topic"]) == 2, a_assignment)
        self.assertTrue(len(b_assignment["test-topic"]) == 2, b_assignment)
        self.assertTrue(len(c_assignment["test-topic"]) == 2, c_assignment)
        self.assertEqual(
            set(
                a_assignment["test-topic"] +
                b_assignment["test-topic"] +
                c_assignment["test-topic"]),
            set(range(6)),
        )
        # and remove one
        a_joined, b_joined = Deferred(), Deferred()

        coords[0].on_join_complete = lambda *args: a_joined.callback(*args)
        coords[1].on_join_complete = lambda *args: b_joined.callback(*args)
        log.warn("removing third")
        yield coords[2].stop()

        log.warn("waiting for a_joined")
        yield a_joined
        log.warn("waiting for b_joined")
        yield b_joined

        log.warn("done")
        yield coords[0].stop()
        yield coords[1].stop()

    @kafka_versions("all")
    @inlineCallbacks
    def test_single_consumergroup_join(self):
        record_stream = DeferredQueue(backlog=1)

        def processor(consumer, records):
            log.debug('processor(%r, %r)', consumer, records)
            record_stream.put(records)

        coord = ConsumerGroup(
            self.client, self.id(),
            topics=[self.topic], processor=processor,
            retry_backoff_ms=100, heartbeat_interval_ms=1000,
            fatal_backoff_ms=3000,
        )
        join_de = self.when_called(coord, 'on_join_complete')
        coord.start()
        self.addCleanup(coord.stop)
        yield join_de

        self.assertIn(self.topic, coord.consumers)
        self.assertEqual(len(coord.consumers[self.topic]), self.num_partitions)
        self.assertEqual(coord.consumers[self.topic][0].topic, self.topic)
        self.assertEqual(coord.consumers[self.topic][0].partition, 0)

        for part in range(self.num_partitions):
            values = yield self.send_messages(part, [part])
            msgs = yield record_stream.get()
            self.assertEqual(msgs[0].partition, part)
            self.assertEqual(msgs[0].message.value, values[0])

    @kafka_versions("all")
    @inlineCallbacks
    def test_two_consumergroup_join(self):
        """
        When a second member joins the consumer group it triggers a rebalance.
        After that completes some partitions are distributed to each member.
        """
        group_id = 'group_for_two'
        self.client2 = KafkaClient(
            self.harness.bootstrap_hosts,
            clientId=self.topic + '2')
        self.addCleanup(self.client2.close)

        record_stream = DeferredQueue(backlog=1)

        def processor(consumer, records):
            log.debug("processor(%r, %r)", consumer, records)
            record_stream.put(records)

        coord = ConsumerGroup(
            self.client, group_id,
            topics=[self.topic], processor=processor,
            retry_backoff_ms=100, heartbeat_interval_ms=1000,
            fatal_backoff_ms=3000,
        )
        de = self.when_called(coord, 'on_join_complete')
        coord_start_d = coord.start()
        self.addCleanup(coord.stop)

        # FIXME: This doesn't seem to get fired reliably.
        coord_start_d
        # self.addCleanup(lambda: coord_start_d)

        yield de

        # send some messages and see that they're processed
        for part in range(self.num_partitions):
            values = yield self.send_messages(part, [part])
            msgs = yield record_stream.get()
            self.assertEqual(msgs[0].partition, part)
            self.assertEqual(msgs[0].message.value, values[0])

        coord2 = ConsumerGroup(
            self.client2, group_id,
            topics=[self.topic], processor=processor,
            retry_backoff_ms=100, heartbeat_interval_ms=1000,
            fatal_backoff_ms=3000,
        )
        de = self.when_called(coord, 'on_join_complete')
        de2 = self.when_called(coord2, 'on_join_complete')
        coord2_start_d = coord2.start()
        self.addCleanup(coord2.stop)

        # FIXME: This doesn't seem to get fired reliably
        coord2_start_d
        # self.addCleanup(lambda: coord2_start_d)

        yield de
        yield de2
        self.assertIn(self.topic, coord.consumers)
        self.assertIn(self.topic, coord2.consumers)
        self.assertEqual(len(coord.consumers[self.topic]), 3)
        self.assertEqual(len(coord2.consumers[self.topic]), 3)
        self.assertNotEqual(
            coord.consumers[self.topic][0].partition,
            coord2.consumers[self.topic][0].partition)

        # after the cluster has re-formed, send some more messages
        # and check that we get them too (and don't get the old messages again)
        for part in range(self.num_partitions):
            values = yield self.send_messages(part, [part])
            msgs = yield record_stream.get()
            self.assertEqual(msgs[0].partition, part)
            self.assertEqual(msgs[0].message.value, values[0])

    @kafka_versions("all")
    @inlineCallbacks
    def test_broker_restart(self):
        """
            restart the kafka broker and verify that the group rejoins
        """
        record_stream = DeferredQueue(backlog=1)

        def processor(consumer, records):
            log.debug('processor(%r, %r)', consumer, records)
            record_stream.put(records)

        coord = ConsumerGroup(
            self.client, self.id(),
            topics=[self.topic], processor=processor,
            retry_backoff_ms=100, heartbeat_interval_ms=1000,
            fatal_backoff_ms=2000,
        )
        join_de = self.when_called(coord, 'on_join_complete')
        coord.start()
        self.addCleanup(coord.stop)

        yield join_de
        self.assertIn(self.topic, coord.consumers)
        self.assertEqual(len(coord.consumers[self.topic]), self.num_partitions)
        self.assertEqual(coord.consumers[self.topic][0].topic, self.topic)
        self.assertEqual(coord.consumers[self.topic][0].partition, 0)

        # restart the broker and see that we re-join and still work
        leave_de = self.when_called(coord, 'on_group_leave')
        prepare_de = self.when_called(coord, 'on_join_prepare')
        join_de = self.when_called(coord, 'on_join_complete')
        self.harness.brokers[0].stop()
        yield leave_de
        self.assertEqual(len(coord.consumers), 0)
        self.harness.brokers[0].restart()
        yield prepare_de
        yield join_de
        self.assertIn(self.topic, coord.consumers)
        self.assertEqual(len(coord.consumers[self.topic]), self.num_partitions)

        for part in range(self.num_partitions):
            values = yield self.send_messages(part, [part])
            msgs = yield record_stream.get()
            self.assertEqual(msgs[0].partition, part)
            self.assertEqual(msgs[0].message.value, values[0])

    @kafka_versions("all")
    @inlineCallbacks
    def test_consumer_rejoin(self):
        """
            trigger a rejoin via consumer commit failure
        """
        group = 'rejoin_group'
        self.client2 = KafkaClient(
            self.harness.bootstrap_hosts,
            clientId=self.topic + '2')
        self.addCleanup(self.client2.close)

        record_stream = DeferredQueue(backlog=1)

        def processor(consumer, records):
            log.debug('processor(%r, %r)', consumer, records)
            record_stream.put(records)

        coord = ConsumerGroup(
            self.client, group,
            topics=[self.topic], processor=processor,
            session_timeout_ms=6000, retry_backoff_ms=100,
            heartbeat_interval_ms=1000, fatal_backoff_ms=3000,
            consumer_kwargs=dict(auto_commit_every_ms=1000),
        )
        coord_start_d = coord.start()
        self.addCleanup(coord.stop)
        # FIXME: This doesn't seem to get fired reliably.
        coord_start_d
        # self.addCleanup(lambda: coord_start_d)

        yield wait_for_assignments(self.topic, self.num_partitions, [coord])

        # kill the heartbeat timer and start joining the second consumer
        while True:
            if coord._heartbeat_looper.running:
                coord._heartbeat_looper.stop()
                break
            else:
                yield async_delay()

        coord2 = ConsumerGroup(
            self.client2, group,
            topics=[self.topic], processor=processor,
            session_timeout_ms=6000, retry_backoff_ms=100,
            heartbeat_interval_ms=1000, fatal_backoff_ms=3000,
            consumer_kwargs=dict(auto_commit_every_ms=1000),
        )
        coord2_start_d = coord2.start()
        self.addCleanup(coord2.stop)
        # FIXME: This doesn't seem to get fired reliably.
        coord2_start_d
        # self.addCleanup(lambda: coord2_start_d)

        # send some messages and see that they're processed
        # the commit will eventually fail because we're rebalancing
        for part in range(15):
            yield async_delay()
            values = yield self.send_messages(part % self.num_partitions, [part])
            msgs = yield record_stream.get()
            if msgs[0].partition != part:
                # once the commit fails, we will see the msg twice
                break
            self.assertEqual(msgs[0].message.value, values[0])

        yield wait_for_assignments(self.topic, self.num_partitions, [coord, coord2])

        # Once assignments have been received we need to ensure that the record
        # stream is clear of any duplicate messages. We do this by producing
        # a sentinel to each partition and consuming messages from the stream
        # until all the sentinels have appeared at least once. At that point
        # any churn should have cleared up and we can depend on lock-step
        # delivery.
        pending_sentinels = {}
        for part in range(self.num_partitions):
            [value] = yield self.send_messages(part, ['sentinel'])
            pending_sentinels[part] = value
        while pending_sentinels:
            [message] = yield record_stream.get()
            if pending_sentinels.get(message.partition) == message.message.value:
                del pending_sentinels[message.partition]

        # after the cluster has re-formed, send some more messages
        # and check that we get them too (and don't get the old messages again)
        record_stream = DeferredQueue(backlog=1)
        for part in range(self.num_partitions):
            yield async_delay()
            [value] = yield self.send_messages(part, [part])
            log.debug('waiting for messages from partition %d', part)
            [message] = yield record_stream.get()
            self.assertEqual(message.partition, part)
            self.assertEqual(message.message.value, value)



# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2017, 2018, 2019 Ciena Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the KafkaCodec class.
"""

from __future__ import absolute_import, division

import struct
from contextlib import contextmanager
from unittest import SkipTest, TestCase

import mock
import six
from mock import sentinel

from .. import kafkacodec
from ..codec import gzip_decode, has_snappy, snappy_decode
from ..common import (
    BrokerMetadata, ChecksumError, ConsumerFetchSizeTooSmall,
    ConsumerMetadataResponse, FetchRequest, FetchResponse, InvalidMessageError,
    Message, OffsetAndMessage, OffsetCommitRequest, OffsetCommitResponse,
    OffsetFetchRequest, OffsetFetchResponse, OffsetRequest, OffsetResponse,
    PartitionMetadata, ProduceRequest, ProduceResponse, ProtocolError,
    TopicMetadata, UnsupportedCodecError, _HeartbeatRequest,
    _HeartbeatResponse, _JoinGroupProtocolMetadata, _JoinGroupRequest,
    _JoinGroupRequestProtocol, _JoinGroupResponse, _JoinGroupResponseMember,
    _LeaveGroupRequest, _LeaveGroupResponse, _SyncGroupMemberAssignment,
    _SyncGroupRequest, _SyncGroupRequestMember, _SyncGroupResponse,
)
from ..kafkacodec import (
    ATTRIBUTE_CODEC_MASK, CODEC_GZIP, CODEC_NONE, CODEC_SNAPPY, KafkaCodec,
    _ReprRequest, create_gzip_message, create_message, create_message_set,
    create_snappy_message,
)
from .testutil import make_send_requests


def create_encoded_metadata_response(broker_data, topic_data):
    encoded = struct.pack('>ii', 3, len(broker_data))
    for node_id, broker in broker_data.items():
        assert isinstance(broker.host, six.string_types), "{!r} must be string".format(broker.host)
        encoded += struct.pack('>ih', node_id, len(broker.host))
        encoded += broker.host.encode('ascii')
        encoded += struct.pack('>i', broker.port)

    encoded += struct.pack('>i', len(topic_data))
    for topic, topic_metadata in topic_data.items():
        assert isinstance(topic, six.string_types), "{!r} must be string".format(topic)
        _, topic_err, partitions = topic_metadata
        encoded += struct.pack('>hh',  topic_err, len(topic))
        encoded += topic.encode('ascii')
        encoded += struct.pack('>i', len(partitions))
        for partition, metadata in partitions.items():
            encoded += struct.pack('>hiii',
                                   metadata.partition_error_code,
                                   partition, metadata.leader,
                                   len(metadata.replicas))
            for replica in metadata.replicas:
                encoded += struct.pack('>i', replica)
            encoded += struct.pack('>i', len(metadata.isr))
            for isr in metadata.isr:
                encoded += struct.pack('>i', isr)

    return encoded


class ReprRequestTests(TestCase):
    def test_produce(self):
        rr = _ReprRequest((
            b'\x00\x00'
            b'\x00\x03'
            b'\x00\x00\x04\x00'
            b'xx'
        ))
        self.assertEqual('ProduceRequest3 correlationId=1024 (10 bytes)', str(rr))

    def test_unknown_key(self):
        rr = _ReprRequest((
            b'\x00\xff'
            b'\x00\x01'
            b'\x00\x00\x00\x00'
        ))
        self.assertEqual('request key=255v1 correlationId=0 (8 bytes)', str(rr))

    def test_too_short(self):
        rr = _ReprRequest(b'\x00\x00\xab\xcd')
        self.assertEqual("invalid request (0000abcd)", str(rr))


class TestKafkaCodec(TestCase):
    def test_create_message(self):
        payload = b"test"
        key = b"key"
        msg = create_message(payload, key)
        self.assertEqual(msg.magic, 0)
        self.assertEqual(msg.attributes, 0)
        self.assertEqual(msg.key, key)
        self.assertEqual(msg.value, payload)

    def test_create_gzip(self):
        message_list = [create_message(b"v1", None),
                        create_message(b"v2", key=b'42')]
        msg = create_gzip_message(message_list)
        self.assertEqual(msg.magic, 0)
        self.assertEqual(msg.attributes, ATTRIBUTE_CODEC_MASK & CODEC_GZIP)
        self.assertEqual(msg.key, None)
        # Need to decode to check since gzipped payload is non-deterministic
        decoded = gzip_decode(msg.value)
        expect = b"".join([
            struct.pack(">q", 0),           # MsgSet offset
            struct.pack(">i", 16),          # MsgSet size
            struct.pack(">i", 1285512130),  # CRC
            struct.pack(">bb", 0, 0),       # Magic, flags
            struct.pack(">i", -1),          # -1 indicates a null key
            struct.pack(">i", 2),           # Msg length (bytes)
            b"v1",                          # Message contents

            struct.pack(">q", 0),           # MsgSet offset
            struct.pack(">i", 18),          # MsgSet size
            struct.pack(">i", 1929437987),  # CRC
            struct.pack(">bb", 0, 0),       # Magic, flags
            struct.pack(">i2s", 2, b'42'),  # Key is 2 bytes long, 42
            struct.pack(">i", 2),           # Msg length (bytes)
            b"v2",                          # Message contents
        ])

        self.assertEqual(decoded, expect)

    def test_create_snappy(self):
        if not has_snappy():
            raise SkipTest("Snappy not available")  # pragma: no cover
        message_list = [create_message(b"v3", key=b'84'),
                        create_message(b"v4", None)]
        msg = create_snappy_message(message_list)
        self.assertEqual(msg.magic, 0)
        self.assertEqual(msg.attributes, ATTRIBUTE_CODEC_MASK & CODEC_SNAPPY)
        self.assertEqual(msg.key, None)
        decoded = snappy_decode(msg.value)
        expect = b"".join([
            struct.pack(">q", 0),           # MsgSet offset
            struct.pack(">i", 18),          # MsgSet size
            struct.pack(">i", 813233088),   # CRC
            struct.pack(">bb", 0, 0),       # Magic, flags
            struct.pack(">i2s", 2, b'84'),  # Key is 2 bytes long, '84'
            struct.pack(">i", 2),           # Msg length (bytes)
            b"v3",                          # Message contents

            struct.pack(">q", 0),           # MsgSet offset
            struct.pack(">i", 16),          # MsgSet size
            struct.pack(">i", 1022734157),  # CRC
            struct.pack(">bb", 0, 0),       # Magic, flags
            struct.pack(">i", -1),          # -1 indicates a null key
            struct.pack(">i", 2),           # Msg length (bytes)
            b"v4",                          # Message contents
        ])

        self.assertEqual(decoded, expect)

    def test_encode_message_header(self):
        expect = b"".join([
            struct.pack(">h", 10),              # API Key
            struct.pack(">h", 0),               # API Version
            struct.pack(">i", 4),               # Correlation Id
            struct.pack(">h", len("client3")),  # Length of clientId
            b"client3",                         # ClientId
        ])

        encoded = KafkaCodec._encode_message_header(b"client3", 4, 10)
        self.assertEqual(encoded, expect)

    def test_encode_message(self):
        message = create_message(b"test", b"key")
        encoded = KafkaCodec._encode_message(message)
        expect = b"".join([
            struct.pack(">i", -1427009701),  # CRC
            struct.pack(">bb", 0, 0),        # Magic, flags
            struct.pack(">i", 3),            # Length of key
            b"key",                          # key
            struct.pack(">i", 4),            # Length of value
            b"test",                         # value
        ])

        self.assertEqual(encoded, expect)

    def test_decode_message(self):
        encoded = b"".join([
            struct.pack(">i", -1427009701),  # CRC
            struct.pack(">bb", 0, 0),        # Magic, flags
            struct.pack(">i", 3),            # Length of key
            b"key",                          # key
            struct.pack(">i", 4),            # Length of value
            b"test",                         # value
        ])

        offset = 10
        (returned_offset, decoded_message) = list(
            KafkaCodec._decode_message(encoded, offset))[0]

        self.assertEqual(returned_offset, offset)
        self.assertEqual(decoded_message, create_message(b"test", b"key"))

    def test_encode_message_failure(self):
        self.assertRaises(ProtocolError,
                          KafkaCodec._encode_message,
                          Message(1, 0, "key", "test"))

    def test_encode_message_set(self):
        message_set = [
            create_message(b"v1", b"k1"),
            create_message(b"v2", b"k2"),
        ]

        encoded = KafkaCodec._encode_message_set(message_set)
        expect = b"".join([
            struct.pack(">q", 0),           # MsgSet Offset
            struct.pack(">i", 18),          # Msg Size
            struct.pack(">i", 1474775406),  # CRC
            struct.pack(">bb", 0, 0),       # Magic, flags
            struct.pack(">i", 2),           # Length of key
            b"k1",                          # Key
            struct.pack(">i", 2),           # Length of value
            b"v1",                          # Value

            struct.pack(">q", 0),          # MsgSet Offset
            struct.pack(">i", 18),         # Msg Size
            struct.pack(">i", -16383415),  # CRC
            struct.pack(">bb", 0, 0),      # Magic, flags
            struct.pack(">i", 2),          # Length of key
            b"k2",                         # Key
            struct.pack(">i", 2),          # Length of value
            b"v2",                         # Value
        ])

        self.assertEqual(encoded, expect)

    def test_decode_message_set(self):
        encoded = b"".join([
            struct.pack(">q", 0),           # MsgSet Offset
            struct.pack(">i", 18),          # Msg Size
            struct.pack(">i", 1474775406),  # CRC
            struct.pack(">bb", 0, 0),       # Magic, flags
            struct.pack(">i", 2),           # Length of key
            b"k1",                          # Key
            struct.pack(">i", 2),           # Length of value
            b"v1",                          # Value

            struct.pack(">q", 1),          # MsgSet Offset
            struct.pack(">i", 18),         # Msg Size
            struct.pack(">i", -16383415),  # CRC
            struct.pack(">bb", 0, 0),      # Magic, flags
            struct.pack(">i", 2),          # Length of key
            b"k2",                         # Key
            struct.pack(">i", 2),          # Length of value
            b"v2",                         # Value
        ])

        msgs = list(KafkaCodec._decode_message_set_iter(encoded))
        self.assertEqual(len(msgs), 2)
        msg1, msg2 = msgs

        returned_offset1, decoded_message1 = msg1
        returned_offset2, decoded_message2 = msg2

        self.assertEqual(returned_offset1, 0)
        self.assertEqual(decoded_message1, create_message(b"v1", b"k1"))

        self.assertEqual(returned_offset2, 1)
        self.assertEqual(decoded_message2, create_message(b"v2", b"k2"))

    def test_decode_message_gzip(self):
        gzip_encoded = (b'\xc0\x11\xb2\xf0\x00\x01\xff\xff\xff\xff\x00\x00\x000'
                        b'\x1f\x8b\x08\x00\xa1\xc1\xc5R\x02\xffc`\x80\x03\x01'
                        b'\x9f\xf9\xd1\x87\x18\x18\xfe\x03\x01\x90\xc7Tf\xc8'
                        b'\x80$wu\x1aW\x05\x92\x9c\x11\x00z\xc0h\x888\x00\x00'
                        b'\x00')
        offset = 11
        messages = list(KafkaCodec._decode_message(gzip_encoded, offset))

        self.assertEqual(len(messages), 2)
        msg1, msg2 = messages

        returned_offset1, decoded_message1 = msg1
        self.assertEqual(returned_offset1, 0)
        self.assertEqual(decoded_message1, create_message(b"v1"))

        returned_offset2, decoded_message2 = msg2
        self.assertEqual(returned_offset2, 0)
        self.assertEqual(decoded_message2, create_message(b"v2"))

    def test_decode_message_snappy(self):
        if not has_snappy():
            raise SkipTest("Snappy not available")  # pragma: no cover
        snappy_encoded = (b'\xec\x80\xa1\x95\x00\x02\xff\xff\xff\xff\x00\x00'
                          b'\x00,8\x00\x00\x19\x01@\x10L\x9f[\xc2\x00\x00\xff'
                          b'\xff\xff\xff\x00\x00\x00\x02v1\x19\x1bD\x00\x10\xd5'
                          b'\x96\nx\x00\x00\xff\xff\xff\xff\x00\x00\x00\x02v2')
        offset = 11
        messages = list(KafkaCodec._decode_message(snappy_encoded, offset))
        self.assertEqual(len(messages), 2)

        msg1, msg2 = messages

        returned_offset1, decoded_message1 = msg1
        self.assertEqual(returned_offset1, 0)
        self.assertEqual(decoded_message1, create_message(b"v1"))

        returned_offset2, decoded_message2 = msg2
        self.assertEqual(returned_offset2, 0)
        self.assertEqual(decoded_message2, create_message(b"v2"))

    def test_decode_message_checksum_error(self):
        invalid_encoded_message = b"This is not a valid encoded message"
        iter = KafkaCodec._decode_message(invalid_encoded_message, 0)
        self.assertRaises(ChecksumError, list, iter)

    # NOTE: The error handling in _decode_message_set_iter() is questionable.
    # If it's modified, the next two tests might need to be fixed.
    def test_decode_message_set_fetch_size_too_small(self):
        self.assertRaises(ConsumerFetchSizeTooSmall,
                          list, KafkaCodec._decode_message_set_iter('a'))

    def test_decode_message_set_stop_iteration(self):
        encoded = b"".join([
            struct.pack(">q", 0),           # MsgSet Offset
            struct.pack(">i", 18),          # Msg Size
            struct.pack(">i", 1474775406),  # CRC
            struct.pack(">bb", 0, 0),       # Magic, flags
            struct.pack(">i", 2),           # Length of key
            b"k1",                          # Key
            struct.pack(">i", 2),           # Length of value
            b"v1",                          # Value

            struct.pack(">q", 1),           # MsgSet Offset
            struct.pack(">i", 18),          # Msg Size
            struct.pack(">i", -16383415),   # CRC
            struct.pack(">bb", 0, 0),       # Magic, flags
            struct.pack(">i", 2),           # Length of key
            b"k2",                          # Key
            struct.pack(">i", 2),           # Length of value
            b"v2",                          # Value
            b"@1$%(Y!",                     # Random padding
        ])

        msgs = list(KafkaCodec._decode_message_set_iter(encoded))
        self.assertEqual(len(msgs), 2)
        msg1, msg2 = msgs

        returned_offset1, decoded_message1 = msg1
        returned_offset2, decoded_message2 = msg2

        self.assertEqual(returned_offset1, 0)
        self.assertEqual(decoded_message1, create_message(b"v1", b"k1"))

        self.assertEqual(returned_offset2, 1)
        self.assertEqual(decoded_message2, create_message(b"v2", b"k2"))

    def test_get_response_correlation_id(self):
        t1 = b"topic1"
        t2 = b"topic2"
        corrID = 30372
        encoded = struct.pack('>iih%dsiihqihqh%dsiihq' % (len(t1), len(t2)),
                              corrID, 2, len(t1), t1, 2, 0, 0, 10, 1, 1, 20,
                              len(t2), t2, 1, 0, 0, 30)
        self.assertEqual(
            corrID, KafkaCodec.get_response_correlation_id(encoded))

    def test_encode_produce_request(self):
        requests = [
            ProduceRequest("topic1", 0, [
                create_message(b"a"),
                create_message(b"b"),
            ]),
            ProduceRequest(u"topic2", 1, [
                create_message(b"c"),
            ]),
        ]

        msg_a_binary = KafkaCodec._encode_message(create_message(b"a"))
        msg_b_binary = KafkaCodec._encode_message(create_message(b"b"))
        msg_c_binary = KafkaCodec._encode_message(create_message(b"c"))

        header = b"".join([
            struct.pack('>h', 0),        # Msg Header, Message type = Produce
            struct.pack('>h', 0),        # Msg Header, API version
            struct.pack('>i', 2),        # Msg Header, Correlation ID
            struct.pack('>h7s', 7, b"client1"),  # Msg Header, The client ID
            struct.pack('>h', 2),        # Num acks required
            struct.pack('>i', 100),      # Request Timeout
            struct.pack('>i', 2),        # The number of requests
        ])

        total_len = len(msg_a_binary) + len(msg_b_binary)
        topic1 = b"".join([
            struct.pack('>h6s', 6, b'topic1'),         # The topic1
            struct.pack('>i', 1),                      # One message set
            struct.pack('>i', 0),                      # Partition 0
            struct.pack('>i', total_len + 24),         # Size of message set
            struct.pack('>q', 0),                      # No offset specified
            struct.pack('>i', len(msg_a_binary)),      # Length of message
            msg_a_binary,                              # Actual message
            struct.pack('>q', 0),                      # No offset specified
            struct.pack('>i', len(msg_b_binary)),      # Length of message
            msg_b_binary,                              # Actual message
        ])

        topic2 = b"".join([
            struct.pack('>h6s', 6, b'topic2'),          # The topic1
            struct.pack('>i', 1),                       # One message set
            struct.pack('>i', 1),                       # Partition 1
            struct.pack('>i', len(msg_c_binary) + 12),  # Size of message set
            struct.pack('>q', 0),                       # No offset specified
            struct.pack('>i', len(msg_c_binary)),       # Length of message
            msg_c_binary,                               # Actual message
        ])

        expected1 = b"".join([header, topic1, topic2])
        expected2 = b"".join([header, topic2, topic1])

        encoded = KafkaCodec.encode_produce_request(
            b"client1", 2, requests, 2, 100)
        self.assertIn(encoded, [expected1, expected2])

    def test_decode_produce_response(self):
        t1 = "topic1"
        t2 = u"topic2"
        encoded = struct.pack('>iih%dsiihqihqh%dsiihq' % (len(t1), len(t2)),
                              2, 2, len(t1), t1.encode(), 2, 0, 0, 10, 1, 1, 20,
                              len(t2), t2.encode(), 1, 0, 0, 30)
        responses = list(KafkaCodec.decode_produce_response(encoded))
        self.assertEqual(responses,
                         [ProduceResponse(t1, 0, 0, 10),
                          ProduceResponse(t1, 1, 1, 20),
                          ProduceResponse(t2, 0, 0, 30)])

    def test_encode_fetch_request(self):
        requests = [
            FetchRequest("topic1", 0, 10, 1024),
            FetchRequest(u"topic2", 1, 20, 100),
        ]

        header = b"".join([
            struct.pack('>h', 1),       # Msg Header, Message type = Fetch
            struct.pack('>h', 0),               # Msg Header, API version
            struct.pack('>i', 3),               # Msg Header, Correlation ID
            struct.pack('>h7s', 7, b"client1"),  # Msg Header, The client ID
            struct.pack('>i', -1),              # Replica Id
            struct.pack('>i', 2),               # Max wait time
            struct.pack('>i', 100),             # Min bytes
            struct.pack('>i', 2),               # Num requests
        ])

        topic1 = b"".join([
            struct.pack('>h6s', 6, b'topic1'),  # Topic
            struct.pack('>i', 1),              # Num Payloads
            struct.pack('>i', 0),              # Partition 0
            struct.pack('>q', 10),             # Offset
            struct.pack('>i', 1024),           # Max Bytes
        ])

        topic2 = b"".join([
            struct.pack('>h6s', 6, b'topic2'),  # Topic
            struct.pack('>i', 1),              # Num Payloads
            struct.pack('>i', 1),              # Partition 0
            struct.pack('>q', 20),             # Offset
            struct.pack('>i', 100),            # Max Bytes
        ])

        expected1 = b"".join([header, topic1, topic2])
        expected2 = b"".join([header, topic2, topic1])

        encoded = KafkaCodec.encode_fetch_request(
            b"client1", 3, requests, 2, 100)
        self.assertIn(encoded, [expected1, expected2])

    def test_decode_fetch_response(self):
        t1 = "topic1"
        t2 = u"topic2"
        msgs = [create_message(m) for m in [b"message1", b"hi", b"boo", b"foo", b"so fun!"]]
        ms1 = KafkaCodec._encode_message_set([msgs[0], msgs[1]])
        ms2 = KafkaCodec._encode_message_set([msgs[2]])
        ms3 = KafkaCodec._encode_message_set([msgs[3], msgs[4]])

        encoded = struct.pack('>iih%dsiihqi%dsihqi%dsh%dsiihqi%ds' %
                              (len(t1), len(ms1), len(ms2), len(t2), len(ms3)),
                              4, 2, len(t1), t1.encode(), 2, 0, 0, 10, len(ms1), ms1, 1,
                              1, 20, len(ms2), ms2, len(t2), t2.encode(), 1, 0, 0, 30,
                              len(ms3), ms3)

        responses = list(KafkaCodec.decode_fetch_response(encoded))

        def expand_messages(response):
            return FetchResponse(response.topic, response.partition,
                                 response.error, response.highwaterMark,
                                 list(response.messages))

        expanded_responses = [expand_messages(r) for r in responses]
        expect = [FetchResponse(t1, 0, 0, 10, [OffsetAndMessage(0, msgs[0]),
                                               OffsetAndMessage(0, msgs[1])]),
                  FetchResponse(t1, 1, 1, 20, [OffsetAndMessage(0, msgs[2])]),
                  FetchResponse(t2, 0, 0, 30, [OffsetAndMessage(0, msgs[3]),
                                               OffsetAndMessage(0, msgs[4])])]
        self.assertEqual(expanded_responses, expect)

    def test_encode_metadata_request_no_topics(self):
        expected = b"".join([
            struct.pack('>h', 3),           # API key metadata fetch
            struct.pack('>h', 0),           # API version
            struct.pack('>i', 4),           # Correlation ID
            struct.pack('>h3s', 3, b"cid"),  # The client ID
            struct.pack('>i', 0),           # No topics, give all the data!
        ])

        encoded = KafkaCodec.encode_metadata_request(b"cid", 4)

        self.assertEqual(encoded, expected)

    def test_encode_metadata_request_with_topics(self):
        expected = b"".join([
            struct.pack('>h', 3),           # API key metadata fetch
            struct.pack('>h', 0),           # API version
            struct.pack('>i', 4),           # Correlation ID
            struct.pack('>h3s', 3, b"cid"),  # The client ID
            struct.pack('>i', 2),           # Number of topics in the request
            struct.pack('>h2s', 2, b"t1"),   # Topic "t1"
            struct.pack('>h2s', 2, b"t2"),   # Topic "t2"
        ])

        encoded = KafkaCodec.encode_metadata_request(b"cid", 4, [u"t1", "t2"])

        self.assertEqual(encoded, expected)

    def test_decode_metadata_response(self):
        node_brokers = {
            0: BrokerMetadata(0, "brokers1.afkak.rdio.com", 1000),
            1: BrokerMetadata(1, "brokers1.afkak.rdio.com", 1001),
            3: BrokerMetadata(3, "brokers2.afkak.rdio.com", 1000),
        }

        topic_partitions = {
            "topic1": TopicMetadata(
                'topic1', 0, {
                    0: PartitionMetadata(u"topic1", 0, 0, 1, (0, 2), (2,)),
                    1: PartitionMetadata("topic1", 1, 1, 3, (0, 1), (0, 1)),
                },
            ),
            u"topic2": TopicMetadata(
                u'topic2', 1, {
                    0: PartitionMetadata(u"topic2", 0, 0, 0, (), ()),
                },
            ),
        }
        encoded = create_encoded_metadata_response(
            node_brokers, topic_partitions)
        decoded = KafkaCodec.decode_metadata_response(encoded)
        self.assertEqual(decoded, (node_brokers, topic_partitions))

    def test_decode_metadata_response_failures(self):
        data = b"".join([
            struct.pack(">ii", 0xCAFE, 0xBABE),  # correlation ID & numbrokers
            struct.pack('>i', 1234),             # Node Id
            struct.pack('>h', len(b"hostname")), b"hostname",  # host
            struct.pack('>i', 1025),             # port
        ])
        self.assertRaises(InvalidMessageError,
                          KafkaCodec.decode_metadata_response, data)

    def test_encode_offset_request(self):
        expected = b"".join([
            struct.pack('>h', 2),            # Message type = offset fetch
            struct.pack('>h', 0),            # API version
            struct.pack('>i', 4),            # Correlation ID
            struct.pack('>h3s', 3, b"cid"),  # The client ID
            struct.pack('>i', -1),           # Replica Id
            struct.pack('>i', 0),            # No topic/partitions
        ])

        encoded = KafkaCodec.encode_offset_request(b"cid", 4)

        self.assertEqual(encoded, expected)

    def test_encode_offset_request__no_payload(self):
        expected = b"".join([
            struct.pack('>h', 2),             # Message type = offset fetch
            struct.pack('>h', 0),             # API version
            struct.pack('>i', 4),             # Correlation ID
            struct.pack('>h3s', 3, b"cid"),   # The client ID
            struct.pack('>i', -1),            # Replica Id
            struct.pack('>i', 1),             # Num topics
            struct.pack(">h6s", 6, b"topic1"),  # Topic for the request
            struct.pack(">i", 2),             # Two partitions

            struct.pack(">i", 3),             # Partition 3
            struct.pack(">q", -1),            # No time offset
            struct.pack(">i", 1),             # One offset requested

            struct.pack(">i", 4),             # Partition 3
            struct.pack(">q", -1),            # No time offset
            struct.pack(">i", 1),             # One offset requested
        ])

        encoded = KafkaCodec.encode_offset_request(b"cid", 4, [
            OffsetRequest('topic1', 3, -1, 1),
            OffsetRequest(u'topic1', 4, -1, 1),
        ])

        self.assertEqual(encoded, expected)

    def test_decode_offset_response(self):
        encoded = b"".join([
            struct.pack(">i", 42),            # Correlation ID
            struct.pack(">i", 1),             # One topics
            struct.pack(">h6s", 6, b"topic1"),  # First topic
            struct.pack(">i", 2),             # Two partitions

            struct.pack(">i", 2),             # Partition 2
            struct.pack(">h", 0),             # No error
            struct.pack(">i", 1),             # One offset
            struct.pack(">q", 4),             # Offset 4

            struct.pack(">i", 4),             # Partition 4
            struct.pack(">h", 0),             # No error
            struct.pack(">i", 1),             # One offset
            struct.pack(">q", 8),             # Offset 8
        ])

        results = KafkaCodec.decode_offset_response(encoded)
        self.assertEqual(set(results), set([
            OffsetResponse(topic='topic1', partition=2, error=0, offsets=(4,)),
            OffsetResponse(topic=u'topic1', partition=4, error=0, offsets=(8,)),
        ]))

    def test_encode_offset_commit_request(self):
        header = b"".join([
            struct.pack('>h', 8),                # Message type = offset commit
            struct.pack('>h', 1),                # API version
            struct.pack('>i', 42),               # Correlation ID
            struct.pack('>h9s', 9, b"client_id"),  # The client ID
            struct.pack('>h8s', 8, b"group_id"),  # The group to commit for
            struct.pack('>i', 996),              # Group generation ID
            struct.pack('>h11s', 11, b'consumer_id'),  # Consumer ID
            struct.pack('>i', 2),                # Num topics
        ])

        topic1 = b"".join([
            struct.pack(">h6s", 6, b"topic1"),    # Topic for the request
            struct.pack(">i", 2),                # Two partitions
            struct.pack(">i", 0),                # Partition 0
            struct.pack(">q", 123),              # Offset 123
            struct.pack(">q", 1437585816816),    # Timestamp in ms > epoch
            struct.pack(">h", -1),               # Null metadata
            struct.pack(">i", 1),                # Partition 1
            struct.pack(">q", 234),              # Offset 234
            struct.pack(">q", 1436981054199),    # Timestamp in ms > epoch
            struct.pack(">h11s", 11, b'My_Metadata'),  # Null metadata
        ])

        topic2 = b"".join([
            struct.pack(">h6s", 6, b"topic2"),    # Topic for the request
            struct.pack(">i", 1),                # One partition
            struct.pack(">i", 2),                # Partition 2
            struct.pack(">q", 345),              # Offset 345
            struct.pack(">q", -1),               # Timestamp 'invalid-time'
            struct.pack(">h", -1),               # Null metadata
        ])

        # A dict is used, so we can't predict the order of the topics...
        expected1 = b"".join([header, topic1, topic2])
        expected2 = b"".join([header, topic2, topic1])

        encoded = KafkaCodec.encode_offset_commit_request(
            b"client_id", 42, u"group_id", 996, u'consumer_id', [
                OffsetCommitRequest("topic1", 0, 123, 1437585816816, None),
                OffsetCommitRequest(u"topic1", 1, 234, 1436981054199,
                                    b'My_Metadata'),
                OffsetCommitRequest(u"topic2", 2, 345, -1, None),
            ])

        self.assertIn(encoded, [expected1, expected2])

    def test_decode_offset_commit_response(self):
        encoded = b"".join([
            struct.pack(">i", 42),            # Correlation ID
            struct.pack(">i", 1),             # One topic
            struct.pack(">h6s", 6, b"topic1"),  # First topic
            struct.pack(">i", 2),             # Two partitions

            struct.pack(">i", 2),             # Partition 2
            struct.pack(">h", 0),             # No error

            struct.pack(">i", 4),             # Partition 4
            struct.pack(">h", 0),             # No error
        ])

        results = KafkaCodec.decode_offset_commit_response(encoded)
        self.assertEqual(set(results), set([
            OffsetCommitResponse(topic='topic1', partition=2, error=0),
            OffsetCommitResponse(topic=u'topic1', partition=4, error=0),
        ]))

    def test_encode_offset_fetch_request(self):
        header = b"".join([
            struct.pack('>h', 9),                  # Message type = offset fetch
            struct.pack('>h', 1),                  # API version
            struct.pack('>i', 42),                 # Correlation ID
            struct.pack('>h9s', 9, b"client_id"),  # The client ID
            struct.pack('>h8s', 8, b"group_id"),   # The group to commit for
            struct.pack('>i', 2),                  # Num topics
        ])

        topic1 = b"".join([
            struct.pack(">h6s", 6, b"topic1"),   # Topic for the request
            struct.pack(">i", 2),                # Two partitions
            struct.pack(">i", 0),                # Partition 0
            struct.pack(">i", 1),                # Partition 1
        ])

        topic2 = b"".join([
            struct.pack(">h6s", 6, b"topic2"),   # Topic for the request
            struct.pack(">i", 1),                # One partitions
            struct.pack(">i", 2),                # Partition 2
        ])

        expected1 = b"".join([header, topic1, topic2])
        expected2 = b"".join([header, topic2, topic1])

        encoded = KafkaCodec.encode_offset_fetch_request(
            b"client_id", 42, u"group_id", [
                OffsetFetchRequest("topic1", 0),
                OffsetFetchRequest(u"topic1", 1),
                OffsetFetchRequest("topic2", 2),
            ])

        self.assertIn(encoded, [expected1, expected2])

    def test_decode_offset_fetch_response(self):
        encoded = b"".join([
            struct.pack(">i", 42),              # Correlation ID
            struct.pack(">i", 1),               # One topics
            struct.pack(">h6s", 6, b"topic1"),  # First topic
            struct.pack(">i", 2),               # Two partitions

            struct.pack(">i", 2),             # Partition 2
            struct.pack(">q", 4),             # Offset 4
            struct.pack(">h4s", 4, b"meta"),  # Metadata
            struct.pack(">h", 0),             # No error

            struct.pack(">i", 4),             # Partition 4
            struct.pack(">q", 8),             # Offset 8
            struct.pack(">h4s", 4, b"meta"),  # Metadata
            struct.pack(">h", 0),             # No error
        ])

        results = KafkaCodec.decode_offset_fetch_response(encoded)
        self.assertEqual(set(results), set([
            OffsetFetchResponse(topic=u'topic1', partition=2, offset=4,
                                error=0, metadata=b"meta"),
            OffsetFetchResponse(topic='topic1', partition=4, offset=8,
                                error=0, metadata=b"meta"),
        ]))

    @contextmanager
    def mock_create_message_fns(self):
        p1 = mock.patch.object(kafkacodec, "create_message",
                               return_value=sentinel.message)
        p2 = mock.patch.object(kafkacodec, "create_gzip_message",
                               return_value=sentinel.gzip_message)
        p3 = mock.patch.object(kafkacodec, "create_snappy_message",
                               return_value=sentinel.snappy_message)
        with p1, p2, p3:
            yield

    def test_create_message_set(self):
        messages = [b'1', b'2', b'3']
        reqs = make_send_requests(messages)

        # Default codec is CODEC_NONE. Expect list of regular messages.
        expect = [sentinel.message] * len(messages)
        with self.mock_create_message_fns():
            message_set = create_message_set(reqs)
        self.assertEqual(message_set, expect)

        # CODEC_NONE: Expect list of regular messages.
        expect = [sentinel.message] * len(messages)
        with self.mock_create_message_fns():
            message_set = create_message_set(reqs, CODEC_NONE)
        self.assertEqual(message_set, expect)

        # CODEC_GZIP: Expect list of one gzip-encoded message.
        expect = [sentinel.gzip_message]
        with self.mock_create_message_fns():
            message_set = create_message_set(reqs, CODEC_GZIP)
        self.assertEqual(message_set, expect)

        # CODEC_SNAPPY: Expect list of one snappy-encoded message.
        expect = [sentinel.snappy_message]
        with self.mock_create_message_fns():
            message_set = create_message_set(reqs, CODEC_SNAPPY)
        self.assertEqual(message_set, expect)

        # Unknown codec should raise UnsupportedCodecError.
        self.assertRaises(UnsupportedCodecError,
                          create_message_set, reqs, -1)

    def test_encode_consumer_metadata_request(self):
        expected = b"".join([
            struct.pack('>h', 10),          # API key ConsumerMetadataRequest
            struct.pack('>h', 0),           # API version
            struct.pack('>i', 9),           # Correlation ID
            struct.pack('>h3s', 3, b"cID"),  # The client ID
            struct.pack('>h6s', 6, b"group1"),   # Consumer group 'group1'
        ])

        encoded = KafkaCodec.encode_consumermetadata_request(b"cID", 9, u"group1")

        self.assertEqual(encoded, expected)

    def test_decode_consumermetadata_response(self):
        expected = ConsumerMetadataResponse(0, 5, "theHost", 4242)
        encoded = b"".join([
            struct.pack('>i', 9),  # Correlation ID
            struct.pack('>h', 0),  # Error Code
            struct.pack('>i', 5),  # Coordinator ID
            struct.pack('>h', len(b"theHost")), b"theHost",  # Coordinator Host
            struct.pack('>i', 4242),  # Coordinator port
        ])
        decoded = KafkaCodec.decode_consumermetadata_response(encoded)
        self.assertEqual(decoded, expected)

    def test_encode_join_group_protocol_metadata(self):
        expected = b"".join([
            struct.pack('>h', 1),                    # version 1
            struct.pack('>ih5s', 1, 5, b"topic"),    # one topic
            struct.pack('>i4s', 4, b"data"),         # Metadata
        ])

        encoded = KafkaCodec.encode_join_group_protocol_metadata(
            1,
            ["topic"],
            b"data",
        )
        self.assertEqual(encoded, expected)

    def test_decode_join_group_protocol_metadata(self):
        expected = _JoinGroupProtocolMetadata(1, ["topic"], b"data")

        encoded = b"".join([
            struct.pack('>h', 1),                    # version 1
            struct.pack('>ih5s', 1, 5, b"topic"),    # one topic
            struct.pack('>i4s', 4, b"data"),         # Metadata
        ])
        decoded = KafkaCodec.decode_join_group_protocol_metadata(encoded)
        self.assertEqual(decoded, expected)

    def test_encode_join_group_request(self):
        expected = b"".join([
            struct.pack('>h', 11),          # API key JoinGroupRequest
            struct.pack('>h', 0),           # API version
            struct.pack('>i', 1),           # Correlation ID
            struct.pack('>h3s', 3, b"cID"),      # The client ID
            struct.pack('>h6s', 6, b"group1"),   # Consumer group 'group1'
            struct.pack('>i', 1),                # Timeout
            struct.pack('>h6s', 6, b"member"),   # Member ID
            struct.pack('>h5s', 5, b"proto"),    # Protocol Type
            struct.pack('>i', 2),                # Two protocols

            struct.pack('>h4s', 4, b"name"),     # Name
            struct.pack('>i4s', 4, b"meta"),     # Metadata
            struct.pack('>h5s', 5, b"name2"),    # Name
            struct.pack('>i5s', 5, b"meta2"),    # Metadata
        ])

        encoded = KafkaCodec.encode_join_group_request(
            b"cID", 1,
            _JoinGroupRequest("group1", 1, "member", "proto", [
                _JoinGroupRequestProtocol("name", b"meta"),
                _JoinGroupRequestProtocol("name2", b"meta2"),
            ]),
        )
        self.assertEqual(encoded, expected)

    def test_decode_join_group_response(self):
        expected = _JoinGroupResponse(
            0, 1, "proto",
            "leader", "member", [
                _JoinGroupResponseMember("id1", b"data1"),
                _JoinGroupResponseMember("id2", b"data2"),
            ])

        encoded = b"".join([
            struct.pack('>i', 9),  # Correlation ID
            struct.pack('>h', 0),  # Error Code
            struct.pack('>i', 1),  # Generation ID
            struct.pack('>h5s', 5, b"proto"),   # Group Protocol
            struct.pack('>h6s', 6, b"leader"),  # Leader ID
            struct.pack('>h6s', 6, b"member"),  # Member ID
            struct.pack('>i', 2),             # Two member datas

            struct.pack('>h3s', 3, b"id1"),     # ID
            struct.pack('>i5s', 5, b"data1"),   # Metadata
            struct.pack('>h3s', 3, b"id2"),     # ID
            struct.pack('>i5s', 5, b"data2"),   # Metadata
        ])
        decoded = KafkaCodec.decode_join_group_response(encoded)
        self.assertEqual(decoded, expected)

    def test_encode_heartbeat_request(self):
        expected = b"".join([
            struct.pack('>h', 12),          # API key HeartbeatRequest
            struct.pack('>h', 0),           # API version
            struct.pack('>i', 1),           # Correlation ID
            struct.pack('>h3s', 3, b"cID"),      # The client ID
            struct.pack('>h6s', 6, b"group1"),   # Group
            struct.pack('>i', 1),                # generation ID
            struct.pack('>h6s', 6, b"member"),   # Member ID
        ])

        encoded = KafkaCodec.encode_heartbeat_request(
            b"cID", 1,
            _HeartbeatRequest("group1", 1, "member"),
        )

        self.assertEqual(encoded, expected)

    def test_decode_heartbeat_response(self):
        expected = _HeartbeatResponse(0)

        encoded = b"".join([
            struct.pack('>i', 9),  # Correlation ID
            struct.pack('>h', 0),  # Error Code
        ])
        decoded = KafkaCodec.decode_heartbeat_response(encoded)
        self.assertEqual(decoded, expected)

    def test_encode_sync_group_request(self):
        expected = b"".join([
            struct.pack('>h', 14),               # API key SyncGroupRequest
            struct.pack('>h', 0),                # API version
            struct.pack('>i', 1),                # Correlation ID
            struct.pack('>h3s', 3, b"cID"),      # The client ID
            struct.pack('>h6s', 6, b"group1"),   # Consumer group 'group1'
            struct.pack('>i', 1),                # Generation ID
            struct.pack('>h6s', 6, b"member"),   # Member ID
            struct.pack('>i', 2),                # Two group assignments

            struct.pack('>h4s', 4, b"name"),     # Name
            struct.pack('>i4s', 4, b"meta"),     # Metadata
            struct.pack('>h5s', 5, b"name2"),    # Name
            struct.pack('>i5s', 5, b"meta2"),    # Metadata
        ])

        encoded = KafkaCodec.encode_sync_group_request(
            b"cID", 1,
            _SyncGroupRequest("group1", 1, "member", [
                _SyncGroupRequestMember("name", b"meta"),
                _SyncGroupRequestMember("name2", b"meta2"),
            ]),
        )
        self.assertEqual(encoded, expected)

    def test_decode_sync_group_response(self):
        expected = _SyncGroupResponse(0, b"data")

        encoded = b"".join([
            struct.pack('>i', 9),  # Correlation ID
            struct.pack('>h', 0),  # Error Code
            struct.pack('>i4s', 4, b"data"),   # Metadata
        ])
        decoded = KafkaCodec.decode_sync_group_response(encoded)
        self.assertEqual(decoded, expected)

    def test_encode_sync_group_member_assignment(self):
        expected = b"".join([
            struct.pack('>h', 1),               # version 1
            struct.pack('>i', 1),               # one assignment
            struct.pack('>h5s', 5, b"topic"),   # topic
            struct.pack('>ii', 1, 5),           # one partition: 5
            struct.pack('>i4s', 4, b"data"),    # Metadata
        ])

        encoded = KafkaCodec.encode_sync_group_member_assignment(
            1,
            {"topic": [5]},
            b"data",
        )
        self.assertEqual(encoded, expected)

    def test_decode_sync_group_member_assignment(self):
        expected = _SyncGroupMemberAssignment(0, {"topic": (5,)}, b"data")

        encoded = b"".join([
            struct.pack('>h', 0),               # version 0
            struct.pack('>i', 1),               # one assignment
            struct.pack('>h5s', 5, b"topic"),   # topic
            struct.pack('>ii', 1, 5),           # one partition: 5
            struct.pack('>i4s', 4, b"data"),    # Metadata
        ])
        decoded = KafkaCodec.decode_sync_group_member_assignment(encoded)
        self.assertEqual(decoded, expected)

    def test_decode_sync_group_member_assignment_bad_version(self):
        """
        Afkak only produces a version 0 SyncGroupMemberAssignment, so any other
        version produces a ProtocolError.
        """
        encoded = b"".join([
            struct.pack('>h', 1),               # version 1
            struct.pack('>i', 1),               # one assignment
            struct.pack('>h5s', 5, b"topic"),   # topic
            struct.pack('>ii', 1, 5),           # one partition: 5
            struct.pack('>i4s', 4, b"data"),    # Metadata
        ])
        self.assertRaises(ProtocolError, KafkaCodec.decode_sync_group_member_assignment, encoded)

    def test_encode_leave_group_request(self):
        expected = b"".join([
            struct.pack('>h', 13),               # API key SyncGroupRequest
            struct.pack('>h', 0),                # API version
            struct.pack('>i', 1),                # Correlation ID
            struct.pack('>h3s', 3, b"cID"),      # The client ID
            struct.pack('>h6s', 6, b"group1"),   # Consumer group 'group1'
            struct.pack('>h6s', 6, b"member"),   # Member ID
        ])

        encoded = KafkaCodec.encode_leave_group_request(
            b"cID", 1,
            _LeaveGroupRequest("group1", "member"),
        )
        self.assertEqual(encoded, expected)

    def test_decode_leave_group_response(self):
        expected = _LeaveGroupResponse(0)

        encoded = struct.pack('>ih', 9, 0)  # Correlation ID, error code
        decoded = KafkaCodec.decode_leave_group_response(encoded)
        self.assertEqual(decoded, expected)



# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2017, 2018 Ciena Corporation.

import collections
import struct

from six import string_types, text_type

from .common import BufferUnderflowError

_NULL_SHORT_STRING = struct.pack('>h', -1)


def _buffer_underflow(what, buf, offset, size):
    return BufferUnderflowError((
        "Not enough data to read {what} at offset {offset:,d}: {size:,d} bytes required,"
        " but {available:,d} available."
    ).format(
        what=what,
        offset=offset,
        size=size,
        available=len(buf) - offset,
    ))


def _coerce_topic(topic):
    """
    Ensure that the topic name is text string of a valid length.

    :param topic: Kafka topic name. Valid characters are in the set ``[a-zA-Z0-9._-]``.
    :raises ValueError: when the topic name exceeds 249 bytes
    :raises TypeError: when the topic is not :class:`unicode` or :class:`str`
    """
    if not isinstance(topic, string_types):
        raise TypeError('topic={!r} must be text'.format(topic))
    if not isinstance(topic, text_type):
        topic = topic.decode('ascii')
    if len(topic) < 1:
        raise ValueError('invalid empty topic name')
    if len(topic) > 249:
        raise ValueError('topic={!r} name is too long: {} > 249'.format(
            topic, len(topic)))
    return topic


def _coerce_consumer_group(consumer_group):
    """
    Ensure that the consumer group is a text string.

    :param consumer_group: :class:`bytes` or :class:`str` instance
    :raises TypeError: when `consumer_group` is not :class:`bytes`
        or :class:`str`
    """
    if not isinstance(consumer_group, string_types):
        raise TypeError('consumer_group={!r} must be text'.format(consumer_group))
    if not isinstance(consumer_group, text_type):
        consumer_group = consumer_group.decode('utf-8')
    return consumer_group


def _coerce_client_id(client_id):
    """
    Ensure the provided client ID is a byte string. If a text string is
    provided, it is encoded as UTF-8 bytes.

    :param client_id: :class:`bytes` or :class:`str` instance
    """
    if isinstance(client_id, type(u'')):
        client_id = client_id.encode('utf-8')
    if not isinstance(client_id, bytes):
        raise TypeError('{!r} is not a valid consumer group (must be'
                        ' str or bytes)'.format(client_id))
    return client_id


def write_int_string(s):
    if s is None:
        return struct.pack('>i', -1)
    return struct.pack('>i', len(s)) + s


def write_short_ascii(s):
    """
    Encode a Kafka short string which represents ASCII text.

    :param str s:
        Text string (`str` on Python 3, `str` or `unicode` on Python 2) or
        ``None``. The string will be ASCII-encoded.

    :returns: length-prefixed `bytes`
    :raises:
        `struct.error` for strings longer than 32767 characters
    """
    if s is None:
        return _NULL_SHORT_STRING
    if not isinstance(s, string_types):
        raise TypeError('{!r} is not text'.format(s))
    return write_short_bytes(s.encode('ascii'))


def write_short_text(s):
    """
    Encode a Kafka short string which represents Unicode text.

    :param str s:
        Text string (`str` on Python 3, `str` or `unicode` on Python 2) or
        ``None``. The string will be UTF-8 encoded.

    :returns: length-prefixed `bytes`
    :raises:
        `struct.error` when the UTF-8 encoded form of the string exceeds
        32767 bytes.
    """
    if s is None:
        return _NULL_SHORT_STRING
    if not isinstance(s, string_types):
        raise TypeError('{!r} is not text'.format(s))
    return write_short_bytes(s.encode('utf-8'))


def write_short_bytes(b):
    """
    Encode a Kafka short string which contains arbitrary bytes. A short string
    is limited to 32767 bytes in length by the signed 16-bit length prefix.
    A length prefix of -1 indicates ``null``, represented as ``None`` in
    Python.

    :param bytes b:
        No more than 32767 bytes, or ``None`` for the null encoding.
    :return: length-prefixed `bytes`
    :raises:
        `struct.error` for strings longer than 32767 characters
    """
    if b is None:
        return _NULL_SHORT_STRING
    if not isinstance(b, bytes):
        raise TypeError('{!r} is not bytes'.format(b))
    elif len(b) > 32767:
        raise struct.error(len(b))
    else:
        return struct.pack('>h', len(b)) + b


def read_short_bytes(data, cur):
    if len(data) < cur + 2:
        raise _buffer_underflow('short string length', data, cur, 2)

    (strlen,) = struct.unpack('>h', data[cur:cur + 2])
    if strlen == -1:
        return None, cur + 2

    cur += 2
    if len(data) < cur + strlen:
        raise _buffer_underflow('short string', data, cur, strlen)

    out = data[cur:cur + strlen]
    return out, cur + strlen


def read_short_ascii(data, cur):
    b, cur = read_short_bytes(data, cur)
    return b.decode('ascii'), cur


def read_short_text(data, cur):
    b, cur = read_short_bytes(data, cur)
    return b.decode('utf-8'), cur


def read_int_string(data, cur):
    if len(data) < cur + 4:
        raise _buffer_underflow('long string length', data, cur, 4)

    (strlen,) = struct.unpack('>i', data[cur:cur + 4])
    if strlen == -1:
        return None, cur + 4

    cur += 4
    if len(data) < cur + strlen:
        raise _buffer_underflow('long string', data, cur, strlen)

    out = data[cur:cur + strlen]
    return out, cur + strlen


def relative_unpack(fmt, data, cur):
    size = struct.calcsize(fmt)
    if len(data) < cur + size:
        raise _buffer_underflow(fmt, data, cur, size)

    out = struct.unpack(fmt, data[cur:cur + size])
    return out, cur + size


def group_by_topic_and_partition(tuples):
    out = collections.defaultdict(dict)
    for t in tuples:
        out[t.topic][t.partition] = t
    return out



# -*- coding: utf-8 -*-
# Copyright 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Endpoints for testing
"""

from __future__ import absolute_import, division

from fnmatch import fnmatchcase
from functools import partial
from pprint import pformat
from struct import Struct

import attr
from twisted.internet import defer
from twisted.internet.interfaces import IAddress, IStreamClientEndpoint
from twisted.internet.protocol import connectionDone
from twisted.logger import Logger
from twisted.protocols.basic import Int32StringReceiver
from twisted.test import iosim
from zope.interface import implementer


@implementer(IStreamClientEndpoint)
@attr.s(frozen=True)
class FailureEndpoint(object):
    """
    Immediately fail every connection attempt.
    """
    reactor = attr.ib()
    host = attr.ib()
    port = attr.ib()

    def connect(self, protocolFactory):
        e = IndentationError('failing to connect {}'.format(protocolFactory))
        return defer.fail(e)


@implementer(IStreamClientEndpoint)
@attr.s(frozen=True)
class BlackholeEndpoint(object):
    """
    Black-hole every connection attempt by returning a deferred which never
    fires.
    """
    reactor = attr.ib()
    host = attr.ib()
    port = attr.ib()

    def connect(self, protocolFactory):
        return defer.Deferred()


@implementer(IAddress)
@attr.s(frozen=True)
class _DebugAddress(object):
    host = attr.ib()
    port = attr.ib()


@attr.s(frozen=True, eq=False)
class Connections(object):
    """Externally controllable endpoint factory

    Each time `Connections` is called to generate an endpoint it returns one
    that records each connection attempt and returns an unresolved deferred.

    :ivar calls:
        List of (host, port) tuples, one for each call to construct an
        endpoint.

    :ivar connects:
        List of (host, port, protocolFactory, deferred), one for each call to
        an endpoint ``connect()`` method.
    """

    calls = attr.ib(init=False, default=attr.Factory(list))
    connects = attr.ib(init=False, default=attr.Factory(list))
    _pumps = attr.ib(init=False, repr=False, default=attr.Factory(list))

    def __call__(self, reactor, host, port):
        self.calls.append((host, port))
        return _PuppetEndpoint(partial(self._connect, host, port))

    def _connect(self, host, port, protocolFactory):
        d = defer.Deferred()
        self.connects.append((host, port, protocolFactory, d))
        return d

    def _match(self, host_pattern):
        for index in range(len(self.connects)):
            host = self.connects[index][0]
            if fnmatchcase(host, host_pattern):
                break
        else:
            raise AssertionError('No match for host pattern {!r}. Searched: {}'.format(
                host_pattern, pformat(self.connects)))
        return self.connects.pop(index)

    def accept(self, host_pattern):
        """
        Accept a pending connection request.

        :param str host_pattern:
            :func:`fnmatch.fnmatchcase` pattern against which the connection
            request must match.

        :returns: `KafkaConnection`
        """
        host, port, protocolFactory, d = self._match(host_pattern)
        peerAddress = _DebugAddress(host, port)
        client_protocol = protocolFactory.buildProtocol(peerAddress)
        assert client_protocol is not None
        server_protocol = KafkaBrokerProtocol()
        client_transport = iosim.FakeTransport(client_protocol, isServer=False, peerAddress=peerAddress)
        server_transport = iosim.FakeTransport(server_protocol, isServer=True)
        pump = iosim.connect(server_protocol, server_transport,
                             client_protocol, client_transport)
        self._pumps.append(pump)
        d.callback(client_protocol)
        return KafkaConnection(
            server=server_protocol,
            client=client_protocol,
            pump=pump,
        )

    def fail(self, host_pattern, reason):
        """
        Fail a pending connection request.

        :param str host_pattern:
            :func:`fnmatch.fnmatchcase` pattern against which the connection
            request must match.

        :param reason: `twisted.python.failure.Failure`
        """
        host, port, protocolFactory, d = self._match(host_pattern)
        d.errback(reason)

    def flush(self):
        """
        Flush I/O on all accepted connections.
        """
        for pump in self._pumps:
            pump.flush()


@implementer(IStreamClientEndpoint)
@attr.s(frozen=True)
class _PuppetEndpoint(object):
    """
    Implementation detail of `Connections`.
    """
    connect = attr.ib()


@attr.s(frozen=True, eq=False)
class KafkaConnection(object):
    """
    Controller for a connection accepted by `Connections`
    """
    #: `twisted.test.iosim.FakeTransport` (client mode)
    client = attr.ib()
    #: `twisted.test.iosim.FakeTransport` (server mode)
    server = attr.ib()
    #: `twisted.test.iosim.IOPump` that flushes data between *client* and *server*
    pump = attr.ib()


@attr.s(frozen=True)
class Request(object):
    _protocol = attr.ib(repr=False)
    api_key = attr.ib()
    api_version = attr.ib()
    correlation_id = attr.ib()
    client_id = attr.ib()
    rest = attr.ib()

    _req_header = Struct('>hhih')
    _resp_header = Struct('>i')

    @classmethod
    def from_bytes(cls, protocol, request):
        """
        :param protocol: `KafkaBrokerProtocol` object
        :param bytes request: Kafka request, less length prefix
        """
        size = cls._req_header.size
        api_key, api_version, correlation_id, client_id_len = cls._req_header.unpack(request[:size])
        if client_id_len < 0:  # Nullable string.
            header_end = size
            client_id = None
        else:
            header_end = size + client_id_len
            client_id = request[size:header_end].decode('utf-8')
        return cls(
            protocol,
            api_key,
            api_version,
            correlation_id,
            client_id,
            request[header_end:],
        )

    def respond(self, response):
        r = self._resp_header.pack(self.correlation_id) + response
        self._protocol.sendString(r)


class KafkaBrokerProtocol(Int32StringReceiver):
    """The server side of the Kafka protocol

    Like a real Kafka broker it handles one request at a time.

    :ivar requests:
    """
    _log = Logger()
    MAX_LENGTH = 2 ** 16  # Small limit for testing.
    disconnected = None

    def connectionMade(self):
        self._log.debug("Connected")
        self.requests = defer.DeferredQueue(backlog=1)

    def stringReceived(self, request):
        """
        Handle a request from the broker.
        """
        r = Request.from_bytes(self, request)
        self._log.debug("Received {request}", request=r)
        self.requests.put(r)

    def connectionLost(self, reason=connectionDone):
        self.disconnected = reason
        self._log.debug("Server connection lost: {reason}", reason=reason)

    def expectRequest(self, api_key, api_version, correlation_id):
        """
        Assert that the next request to enter the queue matches the parameters
        and return it.
        """
        def assert_(request):
            expected = (
                api_key,
                api_version,
                correlation_id,
            )
            actual = (
                request.api_key,
                request.api_version,
                request.correlation_id,
            )
            if expected != actual:
                raise AssertionError((
                    "Request {!r} doesn't match expectation of api_key={!r},"
                    " api_version={!r}, correlation_id={!r}"
                ).format(request, api_key, api_version, correlation_id))

            return request

        return self.requests.get().addCallback(assert_)



# -*- coding: utf-8 -*-
# Copyright 2018 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tools for testing logging

    with capture_logging(logger) as records:
        logger.info("hello")

    [record] = records
    assert record.msg == "hello"
"""

import logging
import unittest
from contextlib import contextmanager


class _CaptureHandler(logging.Handler):
    """Log hander that captures logging calls

    :ivar list records:
        All of the :class:`logging.LogRecord` objects emitted to this handler.
    """
    def __init__(self):
        super(_CaptureHandler, self).__init__()
        self.records = []

    def emit(self, record):
        """Record a log record"""
        self.records.append(record)


@contextmanager
def capture_logging(logger, level=logging.DEBUG):
    """Context manager that captures log records

    :param logger: :class:`logging.Logger`. Propagation of messages is disabled
        while the context manager is active.
    :param level:
        Log level assigned to the logger while the context manager is active.
        Defaults to `logging.DEBUG`.

        If you pass `logging.NOTSET` the level will be inherited from the
        global configuration, which may vary depending on the test harness in
        use (e.g., nose meddles with it). This may lead to brittle tests.

    :returns: :class:`list` of :class:`logging.LogRecord` objects
    """
    handler = _CaptureHandler()
    was_propagating = logger.propagate
    was_level = logger.level
    logger.propagate = False
    logger.setLevel(level)
    logger.addHandler(handler)
    try:
        yield handler.records
    finally:
        logger.removeHandler(handler)
        logger.propagate = was_propagating
        logger.setLevel(was_level)


class CaptureLoggingTests(unittest.TestCase):
    """Test :func:`capture_logging()`"""

    logger = logging.getLogger(__name__).getChild('CaptureLoggingTests')

    def test_capture(self):
        """
        Log messages accumulate in the list.
        """
        with capture_logging(self.logger) as records:
            self.logger.info("hello")
            self.logger.debug("world")

        [r1, r2] = records
        assert r1.msg == "hello"
        assert r2.msg == "world"

    def test_propagate(self):
        """
        Propagation is disabled within the context manager and restored after.
        """
        self.logger.propagate = True

        with capture_logging(self.logger):
            self.assertFalse(self.logger.propagate)

        self.assertTrue(self.logger.propagate)

    def test_level(self):
        """
        The logger level is set within the context manager.
        """
        with capture_logging(self.logger, logging.ERROR):
            self.assertEqual(logging.ERROR, self.logger.level)

        self.assertEqual(logging.NOTSET, self.logger.level)



# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2017, 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""KafkaBrokerClient and private _Request classes.

Low level network client for the Apache Kafka Message Broker.
"""

from __future__ import absolute_import

import logging
from collections import OrderedDict
from datetime import datetime
from functools import partial

import attr
from six.moves import reprlib
from twisted.internet.defer import Deferred, fail, maybeDeferred
from twisted.internet.protocol import ClientFactory
from twisted.internet.task import deferLater
from twisted.python.failure import Failure

from ._protocol import KafkaProtocol
from .common import ClientError, DuplicateRequestError
from .kafkacodec import KafkaCodec, _ReprRequest

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

MAX_RECONNECT_DELAY_SECONDS = 15
INIT_DELAY_SECONDS = 0.1


@attr.s(slots=True)
class _RequestState(object):
    """
    Private helper class to hold flags about in-flight requests.

    :ivar int correlationId: Unique-per-connection ID number for the request.

    :ivar bytes request: Serialized Kafka PDU, less the length prefix.

    :ivar bool expectResponse:
        Is a response expected for this request? If not, it will be discarded
        once it has been written to the protocol.

    :ivar datetime queued:
        When was `_KafkaBrokerClient.makeRequest()` called?

    :ivar datetime sent:
        When was this request written to a connection? `None` if that hasn't
        happened yet.  A request that hasn't been written will be discarded
        immediately (never sent) when cancelled.

    :ivar datetime cancelled:
        When was the deferred for this request canceled? `None` if it hasn't
        been. The queue entry for a cancelled request is kept around if it has
        been sent so that the response correlation logic can differentiate
        between responses that were too late vs. responses to requests that
        were never sent.
    """
    correlationId = attr.ib()
    request = attr.ib(repr=False)
    expectResponse = attr.ib()
    d = attr.ib()
    queued = attr.ib()
    sent = attr.ib(default=None)
    cancelled = attr.ib(default=None)


_aLongerRepr = reprlib.Repr()
_aLongerRepr.maxstring = 1024  # Python 2: str is bytes
_aLongerRepr.maxother = 1024  # Python 3: bytes is not str


class _KafkaBrokerClient(ClientFactory):
    """
    The low-level client which handles transport to a single Kafka broker.

    The KafkaBrokerClient object is responsible for maintaining a connection to
    a single Kafka broker, reconnecting as needed, over which it sends requests
    and receives responses. Callers make requests with :py:method:`makeRequest`
    """
    protocol = KafkaProtocol

    # Reduce log spam from twisted
    noisy = False

    def __init__(self, reactor, endpointFactory, brokerMetadata, clientId,
                 retryPolicy):
        """
        Create a client for a specific broker

        The broker client connects to the broker as needed to handle requests.
        Requests are retried when the connection fails before the client
        receives the response. Requests can be cancelled at any time.

        :param reactor: Twisted reactor to use for connecting and when scheduling
                delayed calls.

        :param endpointFactory:
            Callable that accepts (reactor, host, port) as arguments and
            returns an IStreamClientEndpoint to use to connect to the broker.

        :param BrokerMetadata brokerMetadata:
            Broker node ID, host, and port. This may be updated later by
            calling updateMetadata().

        :param str clientId: Identifying string for log messages. NOTE: not the
            ClientId in the RequestMessage PDUs going over the wire.
        """
        self._reactor = reactor  # ReconnectingClientFactory uses self.clock.
        self._endpointFactory = endpointFactory
        self.node_id = brokerMetadata.node_id
        self.host = brokerMetadata.host
        self.port = brokerMetadata.port
        self.clientId = clientId
        self._retryPolicy = retryPolicy

        # No connector until we try to connect
        self.connector = None
        # The protocol object for the current connection
        self.proto = None
        # ordered dict of _Requests, keyed by requestId
        self.requests = OrderedDict()
        # deferred which fires when the close() completes
        self._dDown = None

    def __repr__(self):
        """return a string representing this KafkaBrokerClient."""
        return '_KafkaBrokerClient<node_id={} {}:{} {}>'.format(
            self.node_id,
            self.host,
            self.port,
            'connected' if self.connected() else 'unconnected',
        )

    def updateMetadata(self, new):
        """
        Update the metadata stored for this broker.

        Future connections made to the broker will use the host and port
        defined in the new metadata. Any existing connection is not dropped,
        however.

        :param new:
            :clas:`afkak.common.BrokerMetadata` with the same node ID as the
            current metadata.
        """
        if self.node_id != new.node_id:
            raise ValueError("Broker metadata {!r} doesn't match node_id={}".format(new, self.node_id))

        self.node_id = new.node_id
        self.host = new.host
        self.port = new.port

    def makeRequest(self, correlationId, request, expectResponse=True):
        """
        Send a request to this broker.

        The request may not be sent immediately, but it should be sent soon.
        A connection to the broker is created when the first request is made.

        Each request is sent at most once per broker connection. If the
        connection drops while requests are outstanding then requests are
        resent upon reconnection in the order originally issued.

        This method returns a deferred that fires with the `bytes` of the
        broker response. It may fail in a few ways:

        - With `ClientError` when the `_KafkaBrokerClient` is closed.
        - With `twisted.internet.defer.CancelledError` if its :meth:`cancel()
          <twisted.internet.defer.Deferred.cancel>`_ method is called.
        - With some other exception in the case of bugs.

        Cancelling the deferred _may_ prevent the request from being sent to
        a broker.

        :param int correlationId:
            ID number used to match responses with requests. This must match
            the number embedded in *request*.

        :param bytes request:
            The serialized request PDU (not including the length prefix).

        :parma bool expectResponse:
            Will the Kafka broker send a response? Unsetting this flag changes
            the result:

              - The request is considered sent as soon as it has been enqueued
                for write to a connection. It will never be resent, and will be
                lost if the connection drops before the broker processes it.
              - The returned deferred will fire with `None` instead of
                a response.

        :returns: Deferred that fires when the request has been processed

        :raises DuplicateRequestError: when *correlationId* is reused. This
            represents a programming error on the caller's part.
        """
        if correlationId in self.requests:
            # Id is duplicate to 'in-flight' request. Reject it, as we
            # won't be able to properly deliver the response(s)
            # Note that this won't protect against a client calling us
            # twice with the same ID, but first with expectResponse=False
            # But that's pathological, and the only defense is to track
            # all requestIds sent regardless of whether we expect to see
            # a response, which is effectively a memory leak...
            raise DuplicateRequestError('Reuse of correlationId={}'.format(correlationId))

        # If we've been told to shutdown (close() called) then fail request
        if self._dDown:
            return fail(ClientError("Broker client for node_id={} {}:{} has been closed".format(
                self.node_id, self.host, self.port)))

        # Ok, we are going to save/send it, create a _Request object to track
        self.requests[correlationId] = tReq = _RequestState(
            correlationId,
            request,
            expectResponse,
            d=Deferred(partial(self._cancelRequest, correlationId)),
            queued=datetime.utcfromtimestamp(self._reactor.seconds()),
        )

        # Do we have a connection over which to send the request?
        if self.proto:
            # Send the request
            # TODO: Flow control
            self._sendRequest(tReq)
        # Have we not even started trying to connect yet? Do so now
        elif not self.connector:
            self._connect()
        return tReq.d

    def disconnect(self):
        """
        Disconnect from the Kafka broker.

        This is used to implement disconnection on timeout as a workaround for
        Kafka connections occasionally getting stuck on the server side under
        load. Requests are not cancelled, so they will be retried.
        """
        if self.proto:
            log.debug('%r Disconnecting from %r', self, self.proto.transport.getPeer())
            self.proto.transport.loseConnection()

    def close(self):
        """Permanently dispose of the broker client.

        This terminates any outstanding connection and cancels any pending
        requests.
        """
        log.debug('%r: close() proto=%r connector=%r', self, self.proto, self.connector)
        assert self._dDown is None
        self._dDown = Deferred()

        if self.proto is not None:
            self.proto.transport.loseConnection()
        elif self.connector is not None:
            def connectingFailed(reason):
                """
                Handle the failure resulting from cancellation.

                :reason: a `Failure`, most likely a cancellation error (but
                    that's not guaranteed).
                :returns: `None` to handle the failure
                """
                log.debug('%r: connection attempt has been cancelled: %r', self, reason)
                self._dDown.callback(None)

            self.connector.addErrback(connectingFailed)
            self.connector.cancel()
        else:
            # Fake a cleanly closing connection
            self._dDown.callback(None)

        try:
            raise ClientError("Broker client for node_id={} {}:{} was closed".format(
                self.node_id, self.host, self.port))
        except Exception:
            reason = Failure()
        # Cancel any requests
        while self.requests:
            correlationId, tReq = self.requests.popitem(True)
            if tReq.cancelled is None:
                tReq.d.errback(reason)
        return self._dDown

    def connected(self):
        """Are we connected to a Kafka broker?"""
        return self.proto is not None

    def _connectionLost(self, reason):
        """Called when the protocol connection is lost

        - Log the disconnection.
        - Mark any outstanding requests as unsent so they will be sent when
          a new connection is made.
        - If closing the broker client, mark completion of that process.

        :param reason:
            Failure that indicates the reason for disconnection.
        """
        log.info('%r: Connection closed: %r', self, reason)

        # Reset our proto so we don't try to send to a down connection
        self.proto = None

        # Mark any in-flight requests as unsent, discard cancelled requests.
        for tReq in list(self.requests.values()):
            if tReq.cancelled is not None:
                del self.requests[tReq.correlationId]
            else:
                tReq.sent = None

        if self._dDown:
            self._dDown.callback(None)
        elif self.requests:
            self._connect()

    def handleResponse(self, response):
        """Handle the response string received by KafkaProtocol.

        Ok, we've received the response from the broker. Find the requestId
        in the message, lookup & fire the deferred with the response.
        """
        correlationId = KafkaCodec.get_response_correlation_id(response)
        # Protect against responses coming back we didn't expect
        tReq = self.requests.pop(correlationId, None)
        if tReq is None:
            # The broker sent us a response to a request we didn't make.
            log.error('Unexpected response with correlationId=%d: %s',
                      correlationId, _aLongerRepr.repr(response))
        elif tReq.cancelled is not None:
            now = datetime.utcfromtimestamp(self._reactor.seconds())
            log.debug(
                'Response to %s arrived %s after it was cancelled (%d bytes)',
                _ReprRequest(tReq.request),
                now - tReq.cancelled,
                len(response),
            )
        else:
            tReq.d.callback(response)

    # # Private Methods # #

    def _sendRequest(self, tReq):
        """Send a single request over our protocol to the Kafka broker."""
        try:
            tReq.sent = datetime.utcfromtimestamp(self._reactor.seconds())
            self.proto.sendString(tReq.request)
        except Exception as e:
            log.exception('%r: Failed to send request %r', self, tReq)
            del self.requests[tReq.correlationId]
            tReq.d.errback(e)
        else:
            if not tReq.expectResponse:
                # Once we've sent a request for which we don't expect a reply,
                # we're done, remove it from requests, and fire the deferred
                # with 'None', since there is no reply to be expected
                del self.requests[tReq.correlationId]
                tReq.d.callback(None)

    def _sendQueued(self):
        """Connection just came up, send the unsent requests."""
        for tReq in list(self.requests.values()):  # must copy, may del
            if tReq.sent is None:
                self._sendRequest(tReq)

    def _cancelRequest(self, correlationId, deferred):
        """
        The ``cancel()`` method of a deferred returned by :meth:`makeRequest()`
        was called. If the request hasn't been sent, remove it from the queue.
        Otherwise it is flagged as cancelled so it won't be resent. The
        queue table entry is retained so if a response comes we don't log an
        error.
        """
        tReq = self.requests[correlationId]
        if tReq.sent is not None:
            tReq.cancelled = datetime.utcfromtimestamp(self._reactor.seconds())
        else:
            del self.requests[correlationId]

    def _abortRequest(self, correlationId, reason):
        """
        Remove a request from the queue and fail its deferred.

        :param int correlationId: Request ID

        :param reason: :class:`twisted.python.failure.Failure` instance to
            errback the request deferred with
        """
        tReq = self.requests.pop(correlationId)
        tReq.d.errback(reason)

    def _connect(self):
        """Connect to the Kafka Broker

        This routine will repeatedly try to connect to the broker (with backoff
        according to the retry policy) until it succeeds.
        """
        def tryConnect():
            self.connector = d = maybeDeferred(connect)
            d.addCallback(cbConnect)
            d.addErrback(ebConnect)

        def connect():
            endpoint = self._endpointFactory(self._reactor, self.host, self.port)
            log.debug('%r: connecting with %s', self, endpoint)
            return endpoint.connect(self)

        def cbConnect(proto):
            log.debug('%r: connected to %r', self, proto.transport.getPeer())
            self._failures = 0
            self.connector = None
            self.proto = proto
            if self._dDown:
                proto.transport.loseConnection()
            else:
                self._sendQueued()

        def ebConnect(fail):
            if self._dDown:
                log.debug('%r: breaking connect loop due to %r after close()', self, fail)
                return fail
            self._failures += 1
            delay = self._retryPolicy(self._failures)
            log.debug('%r: failure %d to connect -> %s; retry in %.2f seconds.',
                      self, self._failures, fail.value, delay)

            self.connector = d = deferLater(self._reactor, delay, lambda: None)
            d.addCallback(cbDelayed)

        def cbDelayed(result):
            tryConnect()

        self._failures = 0
        tryConnect()



# -*- coding: utf-8 -*-
# Copyright 2017, 2018, 2019 Ciena Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import logging
import os
import sys
import time
from random import randint

from twisted.internet.defer import inlineCallbacks
from twisted.trial import unittest

from afkak import Consumer, Producer
from afkak.common import OFFSET_EARLIEST
from afkak.consumer import FETCH_BUFFER_SIZE_BYTES
from afkak.test.testutil import async_delay, random_string

from .intutil import IntegrationMixin, kafka_versions, stat

log = logging.getLogger(__name__)

PARTITION_COUNT = 11
MESSAGE_BLOCK_SIZE = 100
PRODUCE_TIME = 20


class TestPerformanceIntegration(IntegrationMixin, unittest.TestCase):
    harness_kw = dict(
        replicas=3,
        partitions=PARTITION_COUNT,
    )

    # Default partition
    partition = 0

    timeout = PRODUCE_TIME * 3 + 5

    if 'TRAVIS' in os.environ:
        skip = "not run on Travis due to flakiness"

    @kafka_versions("all")
    @inlineCallbacks
    def test_throughput(self):
        # Flag to shutdown
        keep_running = True
        # Count of messages sent
        sent_msgs_count = [0]
        total_messages_size = [0]
        # setup MESSAGE_BLOCK_SIZEx1024-ish byte messages to send over and over
        constant_messages = [
            self.msg(s) for s in [
                random_string(1024) for x in range(MESSAGE_BLOCK_SIZE)]]
        large_messages = [
            self.msg(s) for s in [random_string(FETCH_BUFFER_SIZE_BYTES * 3)
                                  for x in range(MESSAGE_BLOCK_SIZE)]
        ]

        constant_messages_size = len(constant_messages[0]) * MESSAGE_BLOCK_SIZE
        large_messages_size = len(large_messages[0]) * MESSAGE_BLOCK_SIZE

        # Create a producer and send some messages
        producer = Producer(self.client)

        # Create consumers (1/partition)
        consumers = [self.consumer(partition=p, fetch_max_wait_time=50)
                     for p in range(PARTITION_COUNT)]

        def log_error(failure):
            log.exception("Failure sending messages: %r", failure)  # pragma: no cover

        def sent_msgs(resps):
            log.info("Messages Sent: %r", resps)
            sent_msgs_count[0] += MESSAGE_BLOCK_SIZE
            return resps

        def send_msgs():
            # randomly, 1/20 of the time, send large messages
            if randint(0, 19):
                messages = constant_messages
                large = ''
                total_messages_size[0] += constant_messages_size
            else:
                messages = large_messages
                large = ' large'
                total_messages_size[0] += large_messages_size

            log.info("Sending: %d%s messages", len(messages), large)
            d = producer.send_messages(self.topic, msgs=messages)
            # As soon as we get a response from the broker, count them
            # and if we're still supposed to, send more
            d.addCallback(sent_msgs)
            if keep_running:
                d.addCallback(lambda _: self.reactor.callLater(0, send_msgs))
                # d.addCallback(lambda _: send_msgs())
            d.addErrback(log_error)

        # Start sending messages, MESSAGE_BLOCK_SIZE at a time, 1K or 384K each
        send_msgs()

        # Start the consumers from the beginning
        fetch_start = time.time()
        start_ds = [consumer.start(OFFSET_EARLIEST) for consumer in consumers]

        # Let them all run for awhile...
        log.info("Waiting %d seconds...", PRODUCE_TIME)
        yield async_delay(PRODUCE_TIME)
        # Tell the producer to stop
        keep_running = False
        # Wait up to PRODUCE_TIME for the consumers to catch up
        log.info("Waiting up to %d seconds for consumers to finish consuming...", PRODUCE_TIME)
        deadline = time.time() + PRODUCE_TIME * 2
        while time.time() < deadline:
            consumed = sum([len(consumer.processor._messages) for consumer in consumers])
            log.debug("Consumed %d messages.", consumed)
            if sent_msgs_count[0] == consumed:
                break
            yield async_delay(1)
        fetch_time = time.time() - fetch_start
        consumed_bytes = sum([c.processor._messages_bytes[0] for c in consumers])

        result_msg = (
            "Sent: {} messages ({:,} total bytes) in ~{} seconds"
            " ({}/sec), Consumed: {} in {:.2f} seconds."
        ).format(sent_msgs_count[0], total_messages_size[0],
                 PRODUCE_TIME, sent_msgs_count[0] / PRODUCE_TIME,
                 consumed, fetch_time)
        # Log the result, and print to stderr to get around nose capture
        log.info(result_msg)
        print("\n\t Performance Data: " + result_msg, file=sys.stderr)
        # And print data as stats
        stat('Production_Time', PRODUCE_TIME)
        stat('Consumption_Time', fetch_time)
        stat('Messages_Produced', sent_msgs_count[0])
        stat('Messages_Consumed', consumed)
        stat('Messages_Bytes_Produced', total_messages_size[0])
        stat('Messages_Bytes_Consumed', consumed_bytes)
        stat('Messages_Produced_Per_Second', sent_msgs_count[0] / PRODUCE_TIME)
        stat('Messages_Consumed_Per_Second', consumed / fetch_time)
        stat('Message_Bytes_Produced_Per_Second', total_messages_size[0] / PRODUCE_TIME)
        stat('Message_Bytes_Consumed_Per_Second', consumed_bytes / fetch_time)

        # Clean up
        log.debug('Stopping producer: %r', producer)
        yield producer.stop()
        log.debug('Stopping consumers: %r', consumers)
        for consumer in consumers:
            consumer.stop()
        for start_d in start_ds:
            self.successResultOf(start_d)
        # make sure we got all the messages we sent
        self.assertEqual(sent_msgs_count[0], sum([len(consumer.processor._messages) for consumer in consumers]))
        # self.fail("Failing so Nose will emit logging.")

    def consumer(self, **kwargs):
        def make_processor():
            def default_message_proccessor(consumer_instance, messages):
                """Default message processing function

                   Strictly for testing.
                   Just adds the messages to its own _messages attr
                """
                log.debug(
                    'Processor for Consumer: %r Got block of %d messages.\n'
                    'Sizes: %r',
                    consumer_instance, len(messages),
                    [len(m.message.value) for m in messages])
                default_message_proccessor._messages.extend(messages)
                default_message_proccessor._messages_bytes[0] += sum(
                    [len(m.message.value) for m in messages])
                return None

            # Setup a list property '_messages' on the processor function
            default_message_proccessor._messages = []
            default_message_proccessor._messages_bytes = [0]

            return default_message_proccessor

        topic = kwargs.pop('topic', self.topic)
        partition = kwargs.pop('partition', self.partition)
        processor = kwargs.pop('processor', make_processor())
        group = kwargs.pop('consumer_group', None)

        return Consumer(self.client, topic, partition, processor, group,
                        **kwargs)



# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test code for Partitioner(object), RoundRobinPartitioner(object),
and HashedPartitioner(object) classes.
"""
from __future__ import absolute_import, division

import logging
from collections import defaultdict
from math import sqrt
from unittest import TestCase

from afkak.partitioner import (
    HashedPartitioner, Partitioner, RoundRobinPartitioner, pure_murmur2,
)

from .testutil import random_string

log = logging.getLogger(__name__)


# Re-implement std so we don't need numpy just for a couple of tests,
# since it wouldn't build in the pypy tox environment
def std(data):
    if not len(data):
        return float('nan')  # pragma: no cover
    mean = sum([float(x) for x in data]) / len(data)
    var = sum([(mean - x)**2 for x in data]) / len(data)
    return sqrt(var)


class TestPartitioner(TestCase):
    def test_constructor(self):
        topic = "TheTestTopic"
        parts = [1, 2, 3, 4, 5, 6]
        p = Partitioner(topic, parts)
        self.assertEqual(topic, p.topic)
        self.assertEqual(parts, p.partitions)

    def test_partition(self):
        parts = [1, 2, 3, 4, 5, 6]
        p = Partitioner(None, parts)

        self.assertRaises(NotImplementedError, p.partition, "key", parts)


class TestRoundRobinPartitioner(TestCase):
    def test_constructor(self):
        parts = [1, 2, 3, 4, 5, 6]
        p = RoundRobinPartitioner(None, parts)
        self.assertEqual(parts, p.partitions)

    def test_repr(self):
        parts = [1, 3, 5, 7]
        p = RoundRobinPartitioner(None, parts)
        self.assertEqual(p.__repr__(),
                         '<RoundRobinPartitioner False:[1, 3, 5, 7]>')

    def test_partition(self):
        parts = [1, 2, 3, 4, 5, 6]
        p = RoundRobinPartitioner(None, parts)

        for part in parts:
            self.assertEqual(part, p.partition("key", parts))

        for part in parts:
            self.assertEqual(part, p.partition("key", parts))

        # Make sure updates work
        parts = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        for part in parts:
            self.assertEqual(part, p.partition("key", parts))

        for part in parts:
            self.assertEqual(part, p.partition("key", parts))

    def test_set_random_start(self):
        RoundRobinPartitioner.set_random_start(True)
        self.assertTrue(RoundRobinPartitioner.randomStart)
        parts = [1, 2, 3, 4, 5, 6, 7, 8, 19, 20, 30]
        # Try a number of times and check the distribution of the start
        firstParts = defaultdict(lambda: 0)
        trycount = 10000
        for _i in range(trycount):
            p1 = RoundRobinPartitioner(None, parts)
            firstParts[p1.partition(None, parts)] += 1

        self.assertLess(std(firstParts.values()), trycount/100)

        RoundRobinPartitioner.set_random_start(False)
        self.assertFalse(RoundRobinPartitioner.randomStart)
        p2 = RoundRobinPartitioner(None, parts)
        self.assertFalse(p2.randomStart)
        for part in parts:
            self.assertEqual(part, p2.partition(None, parts))


class TestHashedPartitioner(TestCase):
    """TestHashedPartitioner

    For all these tests that compare the result, they were performed also
    with a Java program which matches the Kafka Java client to determine the
    values to match against.
    """
    T1 = "T1"
    parts = range(100000)  # big enough have low probability of collision

    def test_key_none(self):
        """
        A key of None results in a TypeError as it doesn't make sense to
        produce unkeyed messages with a hashed partitioner.
        """
        key = None
        p = HashedPartitioner(self.T1, self.parts)
        self.assertRaises(TypeError, p.partition, key, self.parts)

    def test_key_str(self):
        """
        Bytestring keys are hashed directly.
        """
        key = b'The rain in Spain falls mainly on the plain.'
        expected = (2823782121 & 0x7FFFFFFF) % 100000  # Int is from murmur2
        p = HashedPartitioner(self.T1, self.parts)
        part = p.partition(key, self.parts)
        self.assertEqual(expected, part)

    def test_key_unicode(self):
        """
        Text string keys are converted to UTF-8 and hashed to match the Java
        implementation.
        """
        key = u'슬듢芬'
        expected = (3978338664 & 0x7FFFFFFF) % 100000  # Int is from murmur2
        p = HashedPartitioner(self.T1, self.parts)
        part = p.partition(key, self.parts)
        self.assertEqual(expected, part)

    def test_key_integer(self):
        """
        Integer keys produce a TypeError as passing an integer likely
        represents a programming error.
        """
        key = 123456789
        p = HashedPartitioner(self.T1, self.parts)
        self.assertRaises(TypeError, p.partition, key, self.parts)

    def test_key_bytearray(self):
        """
        :class:`bytearray` keys are hashed directly.
        """
        # 15 letters long to hit 'if extrabytes == 3:'
        key = bytearray(b'lasquinceletras')
        expected = (4030895744 & 0x7FFFFFFF) % 100000  # Int is from murmur2
        p = HashedPartitioner(self.T1, self.parts)
        part = p.partition(key, self.parts)
        self.assertEqual(expected, part)

    def test_key_large(self):
        """
        The hashed partitioner does not limit key length.
        """
        key = u''.join([u"Key:{} ".format(i) for i in range(4096)]).encode('ascii')
        expected = 1765856722 % 100000  # Int is from online hash tool
        p = HashedPartitioner(self.T1, self.parts)
        part = p.partition(key, self.parts)
        self.assertEqual(expected, part)

    def test_key_match_java(self):
        """
        Hashing a UUID produces the same result the Java implementation would.
        """
        key = b'cc54d7f5-8508-4302-bc23-c5d16cfb50fd'
        key = key.decode(encoding='UTF-8', errors='strict')
        parts = range(10)
        expected = 4
        p = HashedPartitioner(self.T1, parts)
        part = p.partition(key, parts)
        self.assertEqual(expected, part)

    def test_partition_distribution(self):
        parts = [1, 2, 3, 4, 5]
        p = HashedPartitioner(self.T1, parts)

        # Make sure we have decent distribution
        keycount = 10000
        key_list = []
        part_keycount = defaultdict(lambda: 0)
        key_to_part = {}
        for _i in range(keycount):
            key = random_string(16)
            key_list.append(key)
            part = p.partition(key, parts)
            part_keycount[part] += 1
            key_to_part[key] = part

        self.assertLess(std(part_keycount.values()), keycount / 100)

        # Ensure we get the same partition for the same key
        for key in key_to_part.keys():
            part = p.partition(key, parts)
            self.assertEqual(part, key_to_part[key])


class TestPureMurmur2(TestCase):
    def test_pure_murmur2(self):
        data = [b'', b'testing', b'PEACH!', b'Gorz!',
                b'987654321', b'_!!_', b'CRINOID']
        expect = [275646681, 2291530147, 2546348827, 1742407956,
                  577579727, 2335345241, 3603193626]
        for d, e in zip(data, expect):
            self.assertEqual(e, pure_murmur2(bytearray(d)))

    def test_pure_murmur2_badarg(self):
        # pure_murmur2 wants bytearray, not string
        self.assertRaises(TypeError, pure_murmur2, "BadArg")
